# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009 Yorik van Havre <yorik@uncreated.net>                        *
# *   (c) 2010 Ken Cline <cline@frii.com>                                   *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides utility functions that are used by many Draft Gui Commands.

These functions are used by different command classes in the `DraftTools`
module. We assume that the graphical interface was already loaded
as they operate on selections and graphical properties.
"""
## @package gui_tool_utils
# \ingroup draftguitools
# \brief Provides utility functions that are used by many Draft Gui Commands.

## \addtogroup draftguitools
# @{
import re

import FreeCAD as App
import FreeCADGui as Gui
import WorkingPlane
from draftutils import gui_utils
from draftutils import params
from draftutils import utils
from draftutils.messages import _wrn
from draftutils.translate import translate


# Set modifier keys from the parameter database
MODS = ["shift", "ctrl", "alt"]


def get_mod_constrain_key():
    return MODS[params.get_param("modconstrain")]


def get_mod_snap_key():
    return MODS[params.get_param("modsnap")]


def get_mod_alt_key():
    return MODS[params.get_param("modalt")]


_HINT_MOD_KEYS = [Gui.UserInput.KeyShift, Gui.UserInput.KeyControl, Gui.UserInput.KeyAlt]


# To allows for easy concatenation the _get_hint_* functions
# always return a list (with a single item or an empty list).


def _get_hint_mod_constrain():
    key = _HINT_MOD_KEYS[params.get_param("modconstrain")]
    return [Gui.InputHint(translate("draft", "%1 constrain"), key)]


def _get_hint_mod_snap():
    if params.get_param("alwaysSnap"):
        return []
    key = _HINT_MOD_KEYS[params.get_param("modsnap")]
    return [Gui.InputHint(translate("draft", "%1 snap"), key)]


def _get_hint_xyz_constrain():
    pattern = re.compile("[A-Z]")
    shortcut_x = params.get_param("inCommandShortcutRestrictX").upper()
    shortcut_y = params.get_param("inCommandShortcutRestrictY").upper()
    shortcut_z = params.get_param("inCommandShortcutRestrictZ").upper()
    if (
        pattern.fullmatch(shortcut_x)
        and pattern.fullmatch(shortcut_y)
        and pattern.fullmatch(shortcut_z)
    ):
        key_x = getattr(Gui.UserInput, "Key" + shortcut_x)
        key_y = getattr(Gui.UserInput, "Key" + shortcut_y)
        key_z = getattr(Gui.UserInput, "Key" + shortcut_z)
        return [
            Gui.InputHint(translate("draft", "%1/%2/%3 switch constraint"), key_x, key_y, key_z)
        ]
    return []


def _get_hint_relative():
    pattern = re.compile("[A-Z]")
    shortcut = params.get_param("inCommandShortcutRelative").upper()
    if pattern.fullmatch(shortcut):
        key = getattr(Gui.UserInput, "Key" + shortcut)
        return [Gui.InputHint(translate("draft", "%1 toggle relative"), key)]
    return []


def _get_hint_global():
    pattern = re.compile("[A-Z]")
    shortcut = params.get_param("inCommandShortcutGlobal").upper()
    if pattern.fullmatch(shortcut):
        key = getattr(Gui.UserInput, "Key" + shortcut)
        return [Gui.InputHint(translate("draft", "%1 toggle global"), key)]
    return []


def _get_hint_continue():
    pattern = re.compile("[A-Z]")
    shortcut = params.get_param("inCommandShortcutContinue").upper()
    if pattern.fullmatch(shortcut):
        key = getattr(Gui.UserInput, "Key" + shortcut)
        return [Gui.InputHint(translate("draft", "%1 toggle continue"), key)]
    return []


def format_unit(exp, unit="mm"):
    """Return a formatting string to set a number to the correct unit."""
    return App.Units.Quantity(exp, App.Units.Length).UserString


formatUnit = format_unit


def select_object(arg):
    """Handle the selection of objects depending on buttons pressed.

    This is a scene event handler, to be called from the Draft tools
    when they need to select an object.
    ::
        self.call = self.view.addEventCallback("SoEvent", select_object)

    Parameters
    ----------
    arg: Coin event
        The Coin event received from the 3D view.

        If it is of type Keyboard and the `ESCAPE` key, it runs the `finish`
        method of the active command.

        If Ctrl key is pressed, multiple selection is enabled until the
        button is released.
        Then it runs the `proceed` method of the active command
        to continue with the command's logic.
    """
    if arg["Type"] == "SoKeyboardEvent":
        if arg["Key"] == "ESCAPE":
            App.activeDraftCommand.finish()
            # TODO: this part raises a coin3D warning about scene traversal.
            # It needs to be fixed.
    elif not arg["CtrlDown"] and Gui.Selection.hasSelection():
        App.activeDraftCommand.proceed()


selectObject = select_object


def has_mod(args, mod):
    """Check if args has a specific modifier.

    Parameters
    ----------
    args: Coin event
        The Coin event received from the 3D view.

    mod: str
        A string indicating the modifier, either `'shift'`, `'ctrl'`,
        or `'alt'`.

    Returns
    -------
    bool
        It returns `args["ShiftDown"]`, `args["CtrlDown"]`,
        or `args["AltDown"]`, depending on the passed value of `mod`.
    """
    if mod == "shift":
        return args["ShiftDown"]
    elif mod == "ctrl":
        return args["CtrlDown"]
    elif mod == "alt":
        return args["AltDown"]


hasMod = has_mod


def set_mod(args, mod, state):
    """Set a specific modifier state in args.

    Parameters
    ----------
    args: Coin event
        The Coin event received from the 3D view.

    mod: str
        A string indicating the modifier, either `'shift'`, `'ctrl'`,
        or `'alt'`.

    state: bool
        The boolean value of `state` is assigned to `args["ShiftDown"]`,
        `args["CtrlDown"]`, or `args["AltDown"]`
        depending on `mod`.
    """
    if mod == "shift":
        args["ShiftDown"] = state
    elif mod == "ctrl":
        args["CtrlDown"] = state
    elif mod == "alt":
        args["AltDown"] = state


setMod = set_mod


def get_point(target, args, noTracker=False):
    """Return a constrained 3D point and its original point.

    It is used by the Draft tools.

    Parameters
    ----------
    target: object (class)
        The target object with a `node` attribute. If this is present,
        return the last node, otherwise return `None`.

        In the Draft tools, `target` is essentially the same class
        of the Gui command, that is, `self`. Therefore, this method
        probably makes more sense as a class method.

    args: Coin event
        The Coin event received from the 3D view.

    noTracker: bool, optional
        It defaults to `False`.
        If it is `True`, the tracking line will not be displayed.

    Returns
    -------
    CoinPoint, Base::Vector3, str
        It returns a tuple with some information.

        The first is the Coin point returned by `Snapper.snap`
        or by the `ActiveView`; the second is that same point
        turned into an `App.Vector`,
        and the third is some information of the point
        returned by the `Snapper` or by the `ActiveView`.
    """
    ui = Gui.draftToolBar
    if not ui.mouse:
        return None, None, None

    if target.node:
        last = target.node[-1]
    else:
        last = None

    smod = has_mod(args, get_mod_snap_key())
    cmod = has_mod(args, get_mod_constrain_key())
    point = None

    if hasattr(Gui, "Snapper"):
        point = Gui.Snapper.snap(
            args["Position"], lastpoint=last, active=smod, constrain=cmod, noTracker=noTracker
        )
        info = Gui.Snapper.snapInfo
        mask = Gui.Snapper.affinity
    if not point:
        p = Gui.ActiveDocument.ActiveView.getCursorPos()
        point = Gui.ActiveDocument.ActiveView.getPoint(p)
        info = Gui.ActiveDocument.ActiveView.getObjectInfo(p)
        mask = None

    ctrlPoint = App.Vector(point)
    wp = WorkingPlane.get_working_plane(update=False)
    if target.node:
        if target.featureName == "Rectangle":
            ui.displayPoint(point, target.node[0], plane=wp, mask=mask)
        else:
            ui.displayPoint(point, target.node[-1], plane=wp, mask=mask)
    else:
        ui.displayPoint(point, plane=wp, mask=mask)
    return point, ctrlPoint, info


getPoint = get_point


def set_working_plane_to_object_under_cursor(mouseEvent):
    """Align the working plane to the face under the cursor.

    The working plane is only aligned if it is `'auto'`.

    Parameters
    ----------
    mouseEvent: Coin event
        Coin mouse event.

    Returns
    -------
    App::DocumentObject or None
        The parent object the face belongs to, if alignment occurred, or None.
    """
    objectUnderCursor = gui_utils.get_3d_view().getObjectInfo(
        (mouseEvent["Position"][0], mouseEvent["Position"][1])
    )

    if not objectUnderCursor:
        return None
    if "Face" not in objectUnderCursor["Component"]:
        return None
    wp = WorkingPlane.get_working_plane(update=False)
    if not wp.auto:
        return None

    import Part

    if "ParentObject" in objectUnderCursor:
        obj = objectUnderCursor["ParentObject"]
        sub = objectUnderCursor["SubName"]
    else:
        obj = App.ActiveDocument.getObject(objectUnderCursor["Object"])
        sub = objectUnderCursor["Component"]
    shape = Part.getShape(obj, sub, needSubElement=True, retType=0)

    if wp.align_to_face(shape, _hist_add=False):
        wp.auto = True
        return obj

    return None


setWorkingPlaneToObjectUnderCursor = set_working_plane_to_object_under_cursor


def set_working_plane_to_selected_object():
    """Align the working plane to a preselected face.

    The working plane is only aligned if it is `'auto'`.

    Returns
    -------
    App::DocumentObject or None
        The parent object the face belongs to, if alignment occurred, or None.
    """
    wp = WorkingPlane.get_working_plane(update=False)
    if not wp.auto:
        return None

    sels = Gui.Selection.getSelectionEx("", 0)

    if (
        len(sels) == 1
        and len(sels[0].SubObjects) == 1
        and sels[0].SubObjects[0].ShapeType == "Face"
    ):
        import Part

        shape = Part.getShape(
            sels[0].Object, sels[0].SubElementNames[0], needSubElement=True, retType=0
        )

        if wp.align_to_face(shape, _hist_add=False):
            wp.auto = True
            return sels[0].Object

    return None


setWorkingPlaneToSelectedObject = set_working_plane_to_selected_object


def get_support(mouseEvent=None):
    """ "Align the working plane to a preselected face or the face under the cursor.

    The working plane is only aligned if it is `'auto'`.

    Parameters
    ----------
    mouseEvent: Coin event, optional
        Defaults to `None`.
        Coin mouse event.

    Returns
    -------
    App::DocumentObject or None
        The parent object the face belongs to, if alignment occurred, or None.
    """
    if mouseEvent is None:
        return set_working_plane_to_selected_object()
    return set_working_plane_to_object_under_cursor(mouseEvent)


getSupport = get_support


def redraw_3d_view():
    """Force a redraw of 3D view or do nothing if it fails."""
    try:
        Gui.ActiveDocument.ActiveView.redraw()
    except AttributeError as err:
        _wrn(err)


redraw3DView = redraw_3d_view

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

# Unit test for the Arch module
from bimtests.TestArchRoof import TestArchRoof
from bimtests.TestArchSpace import TestArchSpace
from bimtests.TestArchWall import TestArchWall
from bimtests.TestArchBuildingPart import TestArchBuildingPart
from bimtests.TestArchAxis import TestArchAxis
from bimtests.TestArchStructure import TestArchStructure
from bimtests.TestArchMaterial import TestArchMaterial
from bimtests.TestArchPanel import TestArchPanel
from bimtests.TestArchWindow import TestArchWindow
from bimtests.TestArchStairs import TestArchStairs
from bimtests.TestArchPipe import TestArchPipe
from bimtests.TestArchCurtainWall import TestArchCurtainWall
from bimtests.TestArchProfile import TestArchProfile
from bimtests.TestArchProject import TestArchProject
from bimtests.TestArchSectionPlane import TestArchSectionPlane
from bimtests.TestArchRebar import TestArchRebar
from bimtests.TestArchGrid import TestArchGrid
from bimtests.TestArchFence import TestArchFence
from bimtests.TestArchEquipment import TestArchEquipment
from bimtests.TestArchFrame import TestArchFrame
from bimtests.TestArchReference import TestArchReference
from bimtests.TestArchSchedule import TestArchSchedule
from bimtests.TestArchTruss import TestArchTruss
from bimtests.TestArchComponent import TestArchComponent
from bimtests.TestWebGLExport import TestWebGLExport



# ***************************************************************************
# *   Copyright (c) 2018 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Results FEM unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import unittest
from os.path import join

import FreeCAD

from . import support_utils as testtools
from .support_utils import fcc_print


class TestResult(unittest.TestCase):
    fcc_print("import TestResult")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test

        # new document
        self.document = FreeCAD.newDocument(self.__class__.__name__)

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars
        fcc_print(
            "\n{0}\n{1} run FEM TestResult tests {2}\n{0}".format(100 * "*", 10 * "*", 64 * "*")
        )

    # ********************************************************************************************
    def get_stress_values(self):
        # stress at node 4 of CalculiX cantilever face load
        stress = (
            -4.52840e02,  # Sxx
            -1.94075e02,  # Syy
            -1.94075e02,  # Szz
            6.11223e01,  # Sxy
            -2.60754e01,  # Sxz
            6.92759e-05,  # Syz
        )
        return stress

    # ********************************************************************************************
    def test_stress_von_mises(self):
        expected_mises = 283.2082
        from femresult.resulttools import calculate_von_mises as vm

        mises = vm(self.get_stress_values())
        # fcc_print(round(mises, 4))
        self.assertEqual(
            round(mises, 4),
            expected_mises,
            "Calculated von Mises stress is not the expected value.",
        )

    # ********************************************************************************************
    def test_stress_principal_std(self):
        expected_principal = (-178.0076, -194.0749, -468.9075, 145.4499)
        from femresult.resulttools import calculate_principal_stress_std as pr

        prin = pr(self.get_stress_values())
        rounded_prin = (round(prin[0], 4), round(prin[1], 4), round(prin[2], 4), round(prin[3], 4))
        # fcc_print(rounded_prin)
        self.assertEqual(
            rounded_prin,
            expected_principal,
            "Calculated principal stresses are not the expected values.",
        )

    # ********************************************************************************************
    def test_stress_principal_reinforced(self):
        expected_principal = (-178.0076, -194.0749, -468.9075, 145.4499)
        from femresult.resulttools import calculate_principal_stress_reinforced as prrc

        prin = prrc(self.get_stress_values())
        rounded_prin = (round(prin[0], 4), round(prin[1], 4), round(prin[2], 4), round(prin[3], 4))
        # fcc_print(rounded_prin)
        self.assertEqual(
            rounded_prin,
            expected_principal,
            "Calculated principal reinforced stresses are not the expected values.",
        )

    # ********************************************************************************************
    def test_rho(self):
        data = (
            (
                # Case1: Governing Eq.14
                (2.000, -2.000, 5.000, 6.000, -4.000, 2.000),
                (0.02400, 0.00400, 0.01400),
            ),
            (
                # Case2: Governing Eq.10+
                (-3.000, -7.000, 0.000, 6.000, -4.000, 2.000),
                (0.00886, 0.00000, 0.00571),
            ),
            (
                # Case3: Governing Eq.5
                (-1.000, -7.000, 10.000, 0.000, 0.000, 5.000),
                (0.00000, 0.00000, 0.02714),
            ),
            (
                # Case4: Governing Eq.13
                (3.000, 0.000, 10.000, 0.000, 5.000, 0.000),
                (0.01600, 0.00000, 0.03000),
            ),
            (
                # Case5: Governing Eq.11-
                (10.000, 7.000, -3.000, 3.000, 1.000, -2.000),
                (0.02533, 0.02133, 0.00000),
            ),
            (
                # Case6: Governing Eq.14
                (4.000, -7.000, 3.000, 7.000, 0.000, -5.000),
                (0.02200, 0.01000, 0.01600),
            ),
            (
                # Case7: Governing Eq.14
                (8.000, -14.000, 6.000, 14.000, 0.000, -10.000),
                (0.04400, 0.02000, 0.03200),
            ),
            (
                # Case8: Governing Eq.17
                (1.000, 0.000, 3.000, 10.000, -8.000, 7.000),
                (0.02486, 0.01750, 0.01720),
            ),
            (
                # Case9: Governing Eq.13
                (0.000, 0.000, 0.000, 10.000, 8.000, 7.000),
                (0.03600, 0.03400, 0.03000),
            ),
            (
                # Case10: Governing Eq.13
                (15.000, 0.000, 0.000, 0.000, 0.000, 0.000),
                (0.03000, 0.00000, 0.00000),
            ),
            (
                # Case11: Governing Eq.13
                (0.000, 0.000, 0.000, 5.000, 0.000, 0.000),
                (0.01000, 0.01000, 0.00000),
            ),
        )

        from femresult.resulttools import calculate_rho as calrho

        for i, case in enumerate(data):
            res = calrho(case[0], 500)
            rhores = (round(res[0], 5), round(res[1], 5), round(res[2], 5))
            # fcc_print("Case{}: {}".format(i + 1 , rhores))
            self.assertEqual(rhores, case[1], f"Calculated rho are not the expected Case{i + 1}.")

    # ********************************************************************************************
    def test_disp_abs(self):
        expected_dispabs = 87.302986
        # x, y, z at node 4 of CalculiX cantilever face load
        disp_xyz = [FreeCAD.Vector(8.12900e00, 3.38889e-02, -8.69237e01)]
        from femresult.resulttools import calculate_disp_abs as dp

        disp_abs = round(dp(disp_xyz)[0], 6)
        # fcc_print(disp_abs)
        self.assertEqual(
            disp_abs, expected_dispabs, "Calculated displacement abs are not the expected values."
        )



# SPDX-License-Identifier: LGPL-2.1-or-later

#**************************************************************************
#   Copyright (c) 2023 David Carter <dcarter@davidcarter.ca>              *
#                                                                         *
#   This file is part of the FreeCAD CAx development system.              *
#                                                                         *
#   This program is free software; you can redistribute it and/or modify  *
#   it under the terms of the GNU Lesser General Public License (LGPL)    *
#   as published by the Free Software Foundation; either version 2 of     *
#   the License, or (at your option) any later version.                   *
#   for detail see the LICENCE text file.                                 *
#                                                                         *
#   FreeCAD is distributed in the hope that it will be useful,            *
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#   GNU Library General Public License for more details.                  *
#                                                                         *
#   You should have received a copy of the GNU Library General Public     *
#   License along with FreeCAD; if not, write to the Free Software        *
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#   USA                                                                   *
#**************************************************************************

"""
Test module for FreeCAD material cards and APIs
"""

import unittest
import FreeCAD
import Materials

import os

parseQuantity = FreeCAD.Units.parseQuantity

class MaterialCreationTestCases(unittest.TestCase):
    """
    Test class for FreeCAD material creation and APIs
    """

    def setUp(self):
        """ Setup function to initialize test data """
        self.ModelManager = Materials.ModelManager()
        self.MaterialManager = Materials.MaterialManager()
        self.uuids = Materials.UUIDs()

    def tearDown(self):
        try:
            material = self.MaterialManager.getMaterialByPath("Example/Frakenstein.FCMat", "User")
            os.remove(material.LibraryRoot + "/Example/Frakenstein.FCMat")
            try:
                os.rmdir(material.LibraryRoot + "/Example")
            except OSError:
                # Can't remove directories that aren't empty
                pass
        except LookupError:
            pass

    def checkNewMaterial(self, material):
        """ Check the state of a newly created material """
        self.assertEqual(len(material.UUID), 36)
        self.assertEqual(len(material.Name), 0)
        self.assertEqual(len(material.Author), 0)
        self.assertEqual(len(material.License), 0)
        self.assertEqual(len(material.Description), 0)
        self.assertEqual(len(material.URL), 0)
        self.assertEqual(len(material.Reference), 0)
        self.assertEqual(len(material.Parent), 0)
        self.assertEqual(len(material.Tags), 0)

    def getQuantity(self, value):
        quantity = parseQuantity(value)
        quantity.Format = { "NumberFormat" : "g",
                            "Precision" : 6 }
        return quantity

    def testCreateMaterial(self):
        """ Create a material with properties """
        material = Materials.Material()
        self.checkNewMaterial(material)

        material.Name = "Frankenstein"
        material.Author = "Mary Shelley"
        material.License = "CC-BY-3.0"
        material.Description = "The sad story of a boy afraid of fire"
        material.URL = "https://www.example.com"
        material.Reference = "ISBN 978-1673287882"

        # Ensure a valid UUID
        self.assertEqual(len(material.UUID), 36)
        uuid = material.UUID

        self.assertEqual(material.Name, "Frankenstein")
        self.assertEqual(material.Author, "Mary Shelley")
        self.assertEqual(material.License, "CC-BY-3.0")
        self.assertEqual(material.Description, "The sad story of a boy afraid of fire")
        self.assertEqual(material.URL, "https://www.example.com")
        self.assertEqual(material.Reference, "ISBN 978-1673287882")

        self.assertEqual(len(material.PhysicalModels), 0)
        self.assertEqual(len(material.AppearanceModels), 0)

        material.addAppearanceModel(self.uuids.TextureRendering)
        self.assertEqual(len(material.AppearanceModels), 1)
        # TextureRendering inherits BasicRendering
        self.assertTrue(material.hasAppearanceModel(self.uuids.BasicRendering))
        self.assertTrue(material.hasAppearanceModel(self.uuids.TextureRendering))

        # Colors are tuples of 3 (rgb) or 4 (rgba) values between 0 and 1
        # All values are set with strings
        material.setAppearanceValue("DiffuseColor", "(1.0, 1.0, 1.0)")
        material.setAppearanceValue("SpecularColor", "(0, 0, 0, 1.0)")

        self.assertEqual(material.AppearanceProperties["DiffuseColor"], "(1.0, 1.0, 1.0)")
        self.assertEqual(material.getAppearanceValue("DiffuseColor"), "(1.0, 1.0, 1.0)")

        self.assertEqual(material.AppearanceProperties["SpecularColor"], "(0, 0, 0, 1.0)")
        self.assertEqual(material.getAppearanceValue("SpecularColor"), "(0, 0, 0, 1.0)")

        # Properties without a value will return None
        self.assertIsNone(material.getAppearanceValue("AmbientColor"))
        self.assertIsNone(material.getAppearanceValue("EmissiveColor"))
        self.assertIsNone(material.getAppearanceValue("Shininess"))
        self.assertIsNone(material.getAppearanceValue("Transparency"))
        self.assertIsNone(material.getAppearanceValue("TexturePath"))
        self.assertIsNone(material.getAppearanceValue("TextureImage"))
        self.assertIsNone(material.getAppearanceValue("TextureScaling"))

        material.addPhysicalModel(self.uuids.Density)
        self.assertEqual(len(material.PhysicalModels), 1)
        self.assertTrue(material.hasPhysicalModel(self.uuids.Density))

        # Quantity properties require units
        with self.assertRaises(ValueError):
            # Units of mass not  density
            material.setPhysicalValue("Density", "99.9 kg")

        material.setPhysicalValue("Density", "99.9")
        self.assertEqual(material.getPhysicalValue("Density").Format["NumberFormat"], "g")
        self.assertEqual(material.getPhysicalValue("Density").UserString, self.getQuantity("99.90 kg/m^3").UserString)

        material.setPhysicalValue("Density", "99.9 kg/m^3")
        self.assertEqual(material.getPhysicalValue("Density").Format["NumberFormat"], "g")
        self.assertEqual(material.getPhysicalValue("Density").UserString, self.getQuantity("99.90 kg/m^3").UserString)

        # MaterialManager is unaware of the material until it is saved
        #
        # When initially creating the material, setting overwrite=True preserves the UUID. This should not
        # be used when saving after properties have been edited as this could adversely affect other
        # documents or parts using the same material. Setting overwrite=False, or omitting it, will change
        # the UUID. It will also fail if the material file already exists.
        #
        # Similarly, saveAsCopy=True preserves the UUID and should be used carefully. It will save an
        # identical copy of the original but in a different location.
        #
        # The third optional parameter is saveInherited. When set to true it will mark models and properties
        # as inherited without duplicating them. When false, they will be copied as uninherited. Avoid
        # self-inheritance as this creates an invalid model. It will have a different UUID than the original.
        #
        self.MaterialManager.save("User", material, "Example/Frakenstein.FCMat", overwrite=True)

        # Now the UUID is valid
        self.assertEqual(len(material.UUID), 36)
        self.assertEqual(material.UUID, uuid)
        self.assertIn(uuid, self.MaterialManager.Materials)
        self.assertIsNotNone(self.MaterialManager.getMaterialByPath("Example/Frakenstein.FCMat", "User"))
        self.assertIsNotNone(self.MaterialManager.getMaterial(uuid))



# SPDX-License-Identifier: LGPL-2.1-or-later

import FreeCAD
import FreeCADGui
import Path
import Path.Dressup.Array as DressupArray
import Path.Dressup.Utils as PathDressup

from PySide.QtCore import QT_TRANSLATE_NOOP

translate = FreeCAD.Qt.translate


class DressupArrayViewProvider(object):
    def __init__(self, vobj):
        self.attach(vobj)

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def attach(self, vobj):
        self.vobj = vobj
        self.obj = vobj.Object
        self.panel = None

    def claimChildren(self):
        return [self.obj.Base]

    def onDelete(self, vobj, args=None):
        if vobj.Object and vobj.Object.Proxy:
            vobj.Object.Proxy.onDelete(vobj.Object, args)
        return True

    def setEdit(self, vobj, mode=0):
        return True

    def unsetEdit(self, vobj, mode=0):
        pass

    def setupTaskPanel(self, panel):
        pass

    def clearTaskPanel(self):
        pass

    def getIcon(self):
        if getattr(PathDressup.baseOp(self.obj), "Active", True):
            return ":/icons/CAM_Dressup.svg"
        else:
            return ":/icons/CAM_OpActive.svg"


class CommandPathDressupArray:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Dressup",
            "MenuText": QT_TRANSLATE_NOOP("CAM_DressupArray", "Array"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_DressupArray",
                "Creates an array from a selected toolpath",
            ),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):
        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            Path.Log.error(translate("CAM_DressupArray", "Select one toolpath object") + "\n")
            return
        baseObject = selection[0]

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction("Create Path Array Dress-up")
        FreeCADGui.addModule("Path.Dressup.Gui.Array")
        FreeCADGui.doCommand(
            "Path.Dressup.Gui.Array.Create(App.ActiveDocument.%s)" % baseObject.Name
        )
        # FreeCAD.ActiveDocument.commitTransaction()  # Final `commitTransaction()` called via TaskPanel.accept()
        FreeCAD.ActiveDocument.recompute()


def Create(base, name="DressupPathArray"):
    FreeCAD.ActiveDocument.openTransaction("Create an Array dressup")
    obj = DressupArray.Create(base, name)
    obj.ViewObject.Proxy = DressupArrayViewProvider(obj.ViewObject)
    obj.Base.ViewObject.Visibility = False
    FreeCAD.ActiveDocument.commitTransaction()
    return obj


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_DressupArray", CommandPathDressupArray())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 three_d                                            *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with arrays.

One of the functions is used to create a `twisted array` object.
See `draftobjects/twistedarray.py`.

This array was developed in order to build a `twisted bridge` object.

See https://forum.freecad.org/viewtopic.php?f=23&t=49617
"""
## @package geo_arrays
# \ingroup draftgeoutils
# \brief Provides various functions to work with arrays.

import lazy_loader.lazy_loader as lz

import FreeCAD as App
from draftutils.messages import _msg

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def print_places(places, title="Places"):
    """Print a vector with a title."""
    _msg(12 * "-")
    _msg(title)
    for i in places:
        _msg("{}".format(i))


def get_init_values(path, count=6):
    """Set values needed to create the array."""
    norm = App.Vector(0, 0, 1)

    # Currently this works with a sketch that has a single edge.
    # Here we need a more general function to extract all edges from a shape,
    # so that the array uses all of them.
    edge = path.Shape.Edges[0]
    edge_length = edge.Length

    n = max((count - 1), 1)
    step = edge_length / n
    inc = 360 / n

    return norm, edge, step, inc


def get_n_params(edge, number, step, norm):
    """Get the parameters needed in each iteration."""
    parameter = edge.getParameterByLength(number * step)

    v0 = edge.valueAt(parameter)
    tan = edge.tangentAt(parameter).normalize()
    binorm = tan.cross(norm).normalize()
    rot = App.Rotation(binorm, tan, norm)

    return v0, tan, rot


def get_twisted_placements(path, count=15, rot_factor=0.25):
    """Get the placements of the twisted array elements."""
    count = max(count, 1)

    (norm, edge, step, inc) = get_init_values(path, count)

    increment = 0
    places = []
    params = []

    for number in range(count):
        v0, tan, rot = get_n_params(edge, number, step, norm)

        angle = increment * rot_factor
        place = App.Placement(v0, tan, angle)
        place.Rotation = place.Rotation * rot
        places.append(place)

        params.append((v0, tan, angle, rot))

        increment += inc

    return places, params


def get_twisted_array_shape(base, path, count=15, rot_factor=0.25):
    """Get the twisted array shape as a compound."""
    places, _ = get_twisted_placements(path, count=count, rot_factor=rot_factor)
    shapes, _ = create_frames(base, places)

    shape = Part.makeCompound(shapes)
    return shape


def get_twisted_bridge_shape(base, path, count=15, rot_factor=0.25, width=100, thickness=10):
    """Get the twisted bridge array shape as a compound."""
    compound = list()

    places, _ = get_twisted_placements(path, count=count, rot_factor=rot_factor)
    # print_places(places)

    shapes, profiles = create_frames(base, places)
    compound.extend(shapes)

    tunnel = make_tunnel(path, profiles)
    compound.append(tunnel)

    # size = base.Shape.Wires[-1].BoundBox.XLength * 0.9
    # size = int(size)
    # thickness = size/12.0

    walkway = make_walkway(path, width, thickness)
    compound.append(walkway)
    shape = Part.makeCompound(compound)

    return shape


def create_frames(obj, places):
    """Create the frames from the placements."""
    len_wires = len(obj.Shape.Wires)
    frames = list()
    profiles = list()

    for i in places:
        frame = obj.Shape.copy()
        frame.Placement = i
        frames.append(frame)
        profiles.append(frame.Wires[len_wires - 1])

    return frames, profiles


def make_tunnel(path, profiles):
    """Create the tunnel shape."""
    edge = path.Shape.Edges[0]
    wire = Part.Wire(edge)
    sweep = wire.makePipeShell(profiles)
    return sweep


def make_walkway(path, width=100, thickness=10):
    """Construct the walkway of the twisted bridge array."""
    spine = path.Shape.Edges[0]

    half_size = width / 2
    offset_height = thickness

    norm1 = App.Vector(0, 0, 1)
    v1 = spine.valueAt(1)
    tan1 = spine.tangentAt(1).normalize()
    binorm1 = tan1.cross(norm1)

    place = App.Placement()
    place.Rotation = App.Rotation(binorm1, norm1, tan1)
    place.move(v1 - App.Vector(0, 0, 3 * offset_height))

    plane = Part.makePlane(width, thickness, App.Vector(-half_size, -2 * offset_height, 0))
    face = Part.Face(plane)
    face.Placement = place.multiply(face.Placement)

    wire = Part.Wire(spine)
    sweep = wire.makePipe(face)

    return sweep


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# (c) 2024 Werner Mayer LGPL

__title__="FreeCAD Python documentation"
__author__ = "Werner Mayer"
__url__ = "https://www.freecad.org"
__doc__ = "Helper module to use pydoc"


import os, sys, pydoc, pkgutil

class FreeCADDoc(pydoc.HTMLDoc):
    def index(self, dir, shadowed=None):
        """ Generate an HTML index for a directory of modules."""
        modpkgs = []
        if shadowed is None: shadowed = {}
        for importer, name, ispkg in pkgutil.iter_modules([dir]):
            if name == 'Init':
                continue
            if name == 'InitGui':
                continue
            if name[-2:] == '_d':
                continue
            modpkgs.append((name, '', ispkg, name in shadowed))
            shadowed[name] = 1

        if len(modpkgs) == 0:
            return None

        modpkgs.sort()
        contents = self.multicolumn(modpkgs, self.modpkglink)
        try:
            return self.bigsection(dir, '#ffffff', '#ee77aa', contents)
        except Exception as e:
            return self.bigsection(dir, 'pkg-content', contents)

def bltinlink(name):
    return '<a href=\"%s.html\">%s</a>' % (name, name)

def getIndexOld():
    pydoc.html = FreeCADDoc()
    title = 'FreeCAD Python Modules Index'

    heading = pydoc.html.heading('<big><big><strong>Python: Index of Modules</strong></big></big>','#ffffff', '#7799ee')

    names = list(filter(lambda x: x != '__main__', sys.builtin_module_names))
    contents = pydoc.html.multicolumn(names, bltinlink)
    indices = ['<p>' + pydoc.html.bigsection('Built-in Modules', '#ffffff', '#ee77aa', contents)]

    names = ['FreeCAD', 'FreeCADGui']
    contents = pydoc.html.multicolumn(names, bltinlink)
    indices.append('<p>' + pydoc.html.bigsection('Built-in FreeCAD Modules', '#ffffff', '#ee77aa', contents))

    seen = {}
    for dir in sys.path:
        dir = os.path.realpath(dir)
        ret = pydoc.html.index(dir, seen)
        if ret != None:
            indices.append(ret)

    contents = heading + ' '.join(indices) + '''<p align=right>
<font color=\"#909090\" face=\"helvetica, arial\"><strong>
pydoc</strong> by Ka-Ping Yee &lt;ping@lfw.org&gt;</font>'''

    htmldocument = pydoc.html.page(title, contents)
    return htmldocument

def getIndexNew():
    pydoc.html = FreeCADDoc()
    title = 'FreeCAD Python Modules Index'

    heading = pydoc.html.heading(
        '<strong class="title">Index of Modules</strong>'
    )

    names = list(filter(lambda x: x != '__main__', sys.builtin_module_names))
    contents = pydoc.html.multicolumn(names, bltinlink)
    indices = ['<p>' + pydoc.html.bigsection('Built-in Modules', 'index', contents)]

    names = ['FreeCAD', 'FreeCADGui']
    contents = pydoc.html.multicolumn(names, bltinlink)
    indices.append('<p>' + pydoc.html.bigsection('Built-in FreeCAD Modules', 'index', contents))

    seen = {}
    for dir in sys.path:
        dir = os.path.realpath(dir)
        ret = pydoc.html.index(dir, seen)
        if ret != None:
            indices.append(ret)

    contents = heading + ' '.join(indices) + '''<p align=right>
<font color=\"#909090\" face=\"helvetica, arial\"><strong>
pydoc</strong> by Ka-Ping Yee &lt;ping@lfw.org&gt;</font>'''

    htmldocument = pydoc.html.page(title, contents)
    return htmldocument

def getIndex():
    try:
        return getIndexOld()
    except Exception as e:
        return getIndexNew()

def getPage(page):
    object, name = pydoc.resolve(page)
    page = pydoc.html.page(pydoc.describe(object), pydoc.html.document(object, name))
    return page



# ***************************************************************************
# *   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import unittest
import math

import FreeCAD
from FreeCAD import Base
from FreeCAD import Units
import Part
import Sketcher
import TestSketcherApp


class TestLoft(unittest.TestCase):
    """Loft Tests"""

    def setUp(self):
        self.Doc = FreeCAD.newDocument("PartDesignTestLoft")

    def testSimpleAdditiveLoftCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.ProfileSketch = self.Doc.addObject("Sketcher::SketchObject", "ProfileSketch")
        self.Body.addObject(self.ProfileSketch)
        TestSketcherApp.CreateRectangleSketch(self.ProfileSketch, (0, 0), (1, 1))
        self.Doc.recompute()
        self.LoftSketch = self.Doc.addObject("Sketcher::SketchObject", "LoftSketch")
        self.Body.addObject(self.LoftSketch)
        self.LoftSketch.MapMode = "FlatFace"
        self.LoftSketch.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        TestSketcherApp.CreateRectangleSketch(self.LoftSketch, (0, 1), (1, 1))
        self.Doc.recompute()
        self.AdditiveLoft = self.Doc.addObject("PartDesign::AdditiveLoft", "AdditiveLoft")
        self.Body.addObject(self.AdditiveLoft)
        self.AdditiveLoft.Profile = self.ProfileSketch
        self.AdditiveLoft.Sections = [self.LoftSketch]
        self.Doc.recompute()
        self.assertAlmostEqual(self.AdditiveLoft.Shape.Volume, 1)

    def testSimpleSubtractiveLoftCase(self):
        self.Body = self.Doc.addObject("PartDesign::Body", "Body")
        self.PadSketch = self.Doc.addObject("Sketcher::SketchObject", "SketchPad")
        self.Body.addObject(self.PadSketch)
        TestSketcherApp.CreateRectangleSketch(self.PadSketch, (0, 0), (1, 1))
        self.Doc.recompute()
        self.Pad = self.Doc.addObject("PartDesign::Pad", "Pad")
        self.Body.addObject(self.Pad)
        self.Pad.Profile = self.PadSketch
        self.Pad.Length = 2
        self.Doc.recompute()
        self.ProfileSketch = self.Doc.addObject("Sketcher::SketchObject", "ProfileSketch")
        self.Body.addObject(self.ProfileSketch)
        TestSketcherApp.CreateRectangleSketch(self.ProfileSketch, (0, 0), (1, 1))
        self.Doc.recompute()
        self.LoftSketch = self.Doc.addObject("Sketcher::SketchObject", "LoftSketch")
        self.Body.addObject(self.LoftSketch)
        self.LoftSketch.MapMode = "FlatFace"
        self.LoftSketch.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        self.Doc.recompute()
        TestSketcherApp.CreateRectangleSketch(self.LoftSketch, (0, 1), (1, 1))
        self.Doc.recompute()
        self.SubtractiveLoft = self.Doc.addObject("PartDesign::SubtractiveLoft", "SubtractiveLoft")
        self.Body.addObject(self.SubtractiveLoft)
        self.SubtractiveLoft.Profile = self.ProfileSketch
        self.SubtractiveLoft.Sections = [self.LoftSketch]
        self.Doc.recompute()
        self.assertAlmostEqual(self.SubtractiveLoft.Shape.Volume, 1)

    def testClosedAdditiveLoftCase(self):
        """Test issue #6156: Loft tool "Closed" option not working"""
        body = self.Doc.addObject("PartDesign::Body", "Body")

        sketch1 = body.newObject("Sketcher::SketchObject", "Sketch")
        sketch1.AttachmentSupport = (self.Doc.XZ_Plane, [""])
        sketch1.MapMode = "FlatFace"
        sketch1.addGeometry(
            Part.Circle(Base.Vector(-40.0, 0.0, 0.0), Base.Vector(0, 0, 1), 10.0), False
        )
        sketch1.addConstraint(Sketcher.Constraint("PointOnObject", 0, 3, -1))
        sketch1.addConstraint(Sketcher.Constraint("Diameter", 0, 20.0))
        sketch1.setDatum(1, Units.Quantity("20.000000 mm"))
        sketch1.addConstraint(Sketcher.Constraint("Distance", -1, 1, 0, 3, 40.0))
        sketch1.setDatum(2, Units.Quantity("40.000000 mm"))

        sketch2 = body.newObject("Sketcher::SketchObject", "Sketch001")
        sketch2.AttachmentSupport = (self.Doc.YZ_Plane, "")
        sketch2.MapMode = "FlatFace"
        sketch2.addGeometry(
            Part.Circle(Base.Vector(-10.0, 0.0, 0.0), Base.Vector(0, 0, 1), 10.0), False
        )
        sketch2.addConstraint(Sketcher.Constraint("PointOnObject", 0, 3, -1))
        sketch2.addConstraint(Sketcher.Constraint("Diameter", 0, 20.0))
        sketch2.setDatum(1, Units.Quantity("20.000000 mm"))
        sketch2.addConstraint(Sketcher.Constraint("Distance", -1, 1, 0, 3, 40.0))
        sketch2.setDatum(2, Units.Quantity("40.000000 mm"))

        sketch3 = body.newObject("Sketcher::SketchObject", "Sketch002")
        sketch3.AttachmentSupport = (self.Doc.getObject("YZ_Plane"), "")
        sketch3.MapMode = "FlatFace"
        sketch3.addGeometry(
            Part.Circle(Base.Vector(40.0, 0.0, 0.0), Base.Vector(0, 0, 1), 10.0), False
        )
        sketch3.addConstraint(Sketcher.Constraint("PointOnObject", 0, 3, -1))
        sketch3.addConstraint(Sketcher.Constraint("Distance", -1, 1, 0, 3, 40.0))
        sketch3.setDatum(1, Units.Quantity("40.000000 mm"))
        sketch3.addConstraint(Sketcher.Constraint("Diameter", 0, 20.0))
        sketch3.setDatum(2, Units.Quantity("20.000000 mm"))

        sketch4 = body.newObject("Sketcher::SketchObject", "Sketch003")
        sketch4.AttachmentSupport = (self.Doc.XZ_Plane, "")
        sketch4.MapMode = "FlatFace"
        sketch4.addGeometry(
            Part.Circle(Base.Vector(40.0, 0.0, 0.0), Base.Vector(0, 0, 1), 10.0), False
        )
        sketch4.addConstraint(Sketcher.Constraint("PointOnObject", 0, 3, -1))
        sketch4.addConstraint(Sketcher.Constraint("Distance", -1, 1, 0, 3, 40.0))
        sketch4.setDatum(1, Units.Quantity("40.000000 mm"))
        sketch4.addConstraint(Sketcher.Constraint("Diameter", 0, 20.0))
        sketch4.setDatum(2, Units.Quantity("20.000000 mm"))

        self.Doc.recompute()

        loft = body.newObject("PartDesign::AdditiveLoft", "AdditiveLoft")
        loft.Profile = sketch1
        loft.Sections = [sketch2, sketch4, sketch3]
        loft.Closed = True

        sketch1.Visibility = False
        sketch2.Visibility = False
        sketch3.Visibility = False
        sketch4.Visibility = False

        self.Doc.recompute()

        self.assertGreater(loft.Shape.Volume, 80000.0)  # 85105.5788704151

    def testPadOnLoftBetweenCones(self):
        """Test issue #15138 adapted from a script by chennes"""
        body = self.Doc.addObject("PartDesign::Body", "Body")
        body.Label = "Body"
        coneBottomSketch = body.newObject("Sketcher::SketchObject", "ConeBottomSketch")
        coneBottomSketch.AttachmentSupport = (self.Doc.getObject("XY_Plane"), [""])
        coneBottomSketch.MapMode = "FlatFace"

        geoList = []
        geoList.append(
            Part.Circle(
                Base.Vector(0.000000, 0.000000, 0.000000),
                Base.Vector(0.000000, 0.000000, 1.000000),
                25.000000,
            )
        )
        coneBottomSketch.addGeometry(geoList, False)
        del geoList

        constraintList = []
        coneBottomSketch.addConstraint(Sketcher.Constraint("Diameter", 0, 25.000000))
        coneBottomSketch.addConstraint(Sketcher.Constraint("Coincident", 0, 3, -1, 1))

        geoList = []
        geoList.append(
            Part.Circle(
                Base.Vector(0.000000, 0.000000, 0.000000),
                Base.Vector(0.000000, 0.000000, 1.000000),
                40.000000,
            )
        )
        coneBottomSketch.addGeometry(geoList, False)
        del geoList

        constraintList = []
        coneBottomSketch.addConstraint(Sketcher.Constraint("Diameter", 1, 40.000000))
        coneBottomSketch.addConstraint(Sketcher.Constraint("Coincident", 1, 3, 0, 3))

        coneTopSketch = body.newObject("Sketcher::SketchObject", "ConeTopSketch")
        coneTopSketch.AttachmentSupport = (self.Doc.getObject("XY_Plane"), [""])
        coneTopSketch.MapMode = "FlatFace"

        geoList = []
        geoList.append(
            Part.Circle(
                Base.Vector(0.000000, 0.000000, 0.000000),
                Base.Vector(0.000000, 0.000000, 1.000000),
                8.000000,
            )
        )
        coneTopSketch.addGeometry(geoList, False)
        del geoList

        constraintList = []
        coneTopSketch.addConstraint(Sketcher.Constraint("Diameter", 0, 8.000000))
        coneTopSketch.addConstraint(Sketcher.Constraint("Coincident", 0, 3, -1, 1))

        geoList = []
        geoList.append(
            Part.Circle(
                Base.Vector(0.000000, 0.000000, 0.000000),
                Base.Vector(0.000000, 0.000000, 1.000000),
                15.000000,
            )
        )
        coneTopSketch.addGeometry(geoList, False)
        del geoList

        constraintList = []
        coneTopSketch.addConstraint(Sketcher.Constraint("Diameter", 1, 15.000000))
        coneTopSketch.addConstraint(Sketcher.Constraint("Coincident", 1, 3, 0, 3))
        coneTopSketch.AttachmentOffset = Base.Placement(
            Base.Vector(0, 0, 20), Base.Rotation(Base.Vector(0, 0, 1), 0)
        )
        self.Doc.recompute()

        cone = body.newObject("PartDesign::AdditiveLoft", "Cone")
        cone.Profile = coneBottomSketch
        cone.Sections = [(coneTopSketch, [""])]
        coneBottomSketch.Visibility = False
        coneTopSketch.Visibility = False
        self.Doc.recompute()

        pad = body.newObject("PartDesign::Pad", "Pad")
        pad.Profile = (
            cone,
            [
                "Face3",
            ],
        )
        pad.Length = 10.000000
        pad.TaperAngle = 0.000000
        pad.UseCustomVector = 0
        pad.Direction = (0, 0, 1)
        pad.ReferenceAxis = None
        pad.AlongSketchNormal = 1
        pad.Type = 0
        pad.UpToFace = None
        pad.Reversed = False
        pad.SideType = "One side"
        pad.Offset = 0
        cone.Visibility = True
        self.Doc.recompute()

        outerConeBottomRadius = 40 / 2
        outerConeTopRadius = 15 / 2
        innerConeBottomRadius = 25 / 2
        innerConeTopRadius = 8 / 2
        coneHeight = 20.0
        padHeight = 10.0
        # Frustum volumes
        outerConeVolume = (
            1.0
            / 3.0
            * math.pi
            * coneHeight
            * (
                outerConeBottomRadius**2
                + outerConeTopRadius**2
                + outerConeBottomRadius * outerConeTopRadius
            )
        )
        innerConeVolume = (
            1.0
            / 3.0
            * math.pi
            * coneHeight
            * (
                innerConeBottomRadius**2
                + innerConeTopRadius**2
                + innerConeBottomRadius * innerConeTopRadius
            )
        )
        frustumVolume = outerConeVolume - innerConeVolume
        topArea = math.pi * (outerConeTopRadius**2) - math.pi * (innerConeTopRadius**2)
        bottomArea = math.pi * (outerConeBottomRadius**2) - math.pi * (innerConeBottomRadius**2)
        padVolume = topArea * padHeight
        self.assertEqual(len(cone.Shape.Faces), 4)  # Bottom, Inner, Top, Outer
        self.assertAlmostEqual(cone.Shape.Faces[2].Area, topArea)
        self.assertAlmostEqual(cone.Shape.Faces[0].Area, bottomArea)
        self.assertAlmostEqual(cone.Shape.Volume, frustumVolume)
        self.assertAlmostEqual(
            pad.Shape.Volume, frustumVolume + padVolume
        )  # contains volume of previous in Body
        self.assertAlmostEqual(
            body.Shape.Volume, frustumVolume + padVolume
        )  # Overall body volume matches

    def testTwoFacesAdditiveLoftCase(self):
        """Test issue #19183: Loft tool "Loft between faces no longer works"""
        body = self.Doc.addObject("PartDesign::Body", "Body")

        sketch1 = body.newObject("Sketcher::SketchObject", "Sketch")

        sketch1.addGeometry(
            Part.LineSegment(
                Base.Vector(-2.060394, -1.332045, 0),
                Base.Vector(-19.922129, -27.589359, 0),
            ),
            False,
        )
        sketch1.addGeometry(
            Part.LineSegment(
                Base.Vector(1.940183, -1.501086, 0),
                Base.Vector(16.928263, -28.265512, 0),
            ),
            False,
        )
        sketch1.addGeometry(
            Part.ArcOfCircle(
                Part.Circle(Base.Vector(0.418837, -1.275699, 0), Base.Vector(0, 0, 1), 32.879378),
                -1.751723,
                -1.454683,
            ),
            False,
        )
        sketch1.addGeometry(
            Part.ArcOfCircle(
                Part.Circle(
                    Base.Vector(-9.385396, -30.124937, 0),
                    Base.Vector(0, 0, 1),
                    8.359959,
                ),
                2.847554,
                4.831818,
            ),
            False,
        )
        sketch1.addGeometry(
            Part.ArcOfCircle(
                Part.Circle(
                    Base.Vector(3.236143, -29.279745, 0),
                    Base.Vector(0, 0, 1),
                    11.449505,
                ),
                -1.076432,
                0.044306,
            ),
            False,
        )
        sketch1.addConstraint(Sketcher.Constraint("Coincident", 0, 1, 1, 1))
        sketch1.addConstraint(Sketcher.Constraint("Coincident", 0, 1, 2, 3))
        sketch1.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))
        sketch1.addConstraint(Sketcher.Constraint("Tangent", 0, 2, 3, 1))
        sketch1.addConstraint(Sketcher.Constraint("Tangent", 1, 2, 4, 2))
        sketch1.addConstraint(Sketcher.Constraint("Tangent", 2, 1, 3, 2))
        sketch1.addConstraint(Sketcher.Constraint("Tangent", 2, 2, 4, 1))
        sketch1.addConstraint(Sketcher.Constraint("Symmetric", 0, 2, 1, 2, -2))
        sketch1.addConstraint(Sketcher.Constraint("Symmetric", 2, 1, 2, 2, -2))
        sketch1.delConstraint(8)
        sketch1.addConstraint(Sketcher.Constraint("Radius", 2, 39.936694))
        sketch1.setDatum(8, Units.Quantity("40.000000 mm"))
        sketch1.addConstraint(Sketcher.Constraint("DistanceX", 2, 1, 2, 2, 16.771341))
        sketch1.setDatum(9, Units.Quantity("10.000000 mm"))
        sketch1.addConstraint(Sketcher.Constraint("Distance", 0, 30.965710))
        sketch1.setDatum(10, Units.Quantity("30.000000 mm"))

        sketch2 = body.newObject("Sketcher::SketchObject", "Sketch001")
        sketch2.addGeometry(
            Part.LineSegment(
                Base.Vector(-2.060394, -1.332045, 0),
                Base.Vector(-19.922129, -27.589359, 0),
            ),
            False,
        )
        sketch2.addGeometry(
            Part.LineSegment(
                Base.Vector(1.940183, -1.501086, 0),
                Base.Vector(16.928263, -28.265512, 0),
            ),
            False,
        )
        sketch2.addGeometry(
            Part.ArcOfCircle(
                Part.Circle(Base.Vector(0.418837, -1.275699, 0), Base.Vector(0, 0, 1), 32.879378),
                -1.751723,
                -1.454683,
            ),
            False,
        )
        sketch2.addGeometry(
            Part.ArcOfCircle(
                Part.Circle(
                    Base.Vector(-9.385396, -30.124937, 0),
                    Base.Vector(0, 0, 1),
                    8.359959,
                ),
                2.847554,
                4.831818,
            ),
            False,
        )
        sketch2.addGeometry(
            Part.ArcOfCircle(
                Part.Circle(
                    Base.Vector(3.236143, -29.279745, 0),
                    Base.Vector(0, 0, 1),
                    11.449505,
                ),
                -1.076432,
                0.044306,
            ),
            False,
        )
        sketch2.addConstraint(Sketcher.Constraint("Coincident", 0, 1, 1, 1))
        sketch2.addConstraint(Sketcher.Constraint("Coincident", 0, 1, 2, 3))
        sketch2.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))
        sketch2.addConstraint(Sketcher.Constraint("Tangent", 0, 2, 3, 1))
        sketch2.addConstraint(Sketcher.Constraint("Tangent", 1, 2, 4, 2))
        sketch2.addConstraint(Sketcher.Constraint("Tangent", 2, 1, 3, 2))
        sketch2.addConstraint(Sketcher.Constraint("Tangent", 2, 2, 4, 1))
        sketch2.addConstraint(Sketcher.Constraint("Symmetric", 0, 2, 1, 2, -2))
        sketch2.addConstraint(Sketcher.Constraint("Symmetric", 2, 1, 2, 2, -2))
        sketch2.delConstraint(8)
        sketch2.addConstraint(Sketcher.Constraint("Radius", 2, 39.936694))
        sketch2.setDatum(8, Units.Quantity("30.000000 mm"))
        sketch2.addConstraint(Sketcher.Constraint("Radius", 3, 1.020288))
        sketch2.setDatum(9, Units.Quantity("3.600000 mm"))
        sketch2.addConstraint(Sketcher.Constraint("DistanceX", 2, 1, 2, 2, 10.926759))
        sketch2.setDatum(10, Units.Quantity("10.000000 mm"))
        self.Doc.recompute()

        pad1 = body.newObject("PartDesign::Pad", "Pad")

        pad1.Profile = sketch1
        pad1.Length = 10.000000
        pad1.TaperAngle = 0.000000
        pad1.Reversed = 1
        self.Doc.recompute()
        sketch1.Visibility = False

        pad2 = body.newObject("PartDesign::Pad", "Pad001")

        pad2.Profile = sketch2
        pad2.Length = 10.000000
        pad2.TaperAngle = 0.000000
        self.Doc.recompute()
        sketch2.Visibility = False
        pad1.Visibility = False
        pad2.Visibility = True
        body.Tip = pad2
        self.assertGreater(pad2.Shape.Volume, 8715.0)  # 8720.024151557787 pre-Loft

        loft = body.newObject("PartDesign::AdditiveLoft", "AdditiveLoft")
        loft.Profile = (
            self.Doc.getObject("Pad001"),
            [
                "Face7",
            ],
        )
        loft.Sections = [
            (
                self.Doc.getObject("Pad001"),
                [
                    "Face10",
                ],
            )
        ]
        loft.Closed = False

        self.Doc.recompute()
        body.Tip = loft
        self.assertGreater(body.Shape.Volume, 9220.0)  # 9221.776241582389 post-Loft
        self.Doc.recompute()

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("PartDesignTestLoft")
        # print ("omit closing document for debugging")



# ***************************************************************************
# *   Copyright (c) 2007 Juergen Riegel <juergen.riegel@web.de>             *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

# Open and edit only in UTF-8 !!!!!!

import FreeCAD, os, unittest, tempfile


# ---------------------------------------------------------------------------
# define the functions to test the FreeCAD Document code
# ---------------------------------------------------------------------------


class UnicodeBasicCases(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("CreateTest")

    def testUnicodeLabel(self):
        L1 = self.Doc.addObject("App::FeatureTest", "Label_1")
        L1.Label = "हिन्दी"
        self.assertTrue(L1.Label == "हिन्दी")

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("CreateTest")


class DocumentSaveRestoreCases(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("SaveRestoreTests")
        L1 = self.Doc.addObject("App::FeatureTest", "Label_1")
        L1.Label = "हिन्दी"
        self.TempPath = tempfile.gettempdir()
        FreeCAD.Console.PrintLog("  Using temp path: " + self.TempPath + "\n")

    def testSaveAndRestore(self):
        # saving and restoring
        SaveName = self.TempPath + os.sep + "UnicodeTest.FCStd"
        self.Doc.saveAs(SaveName)
        FreeCAD.closeDocument("SaveRestoreTests")
        self.Doc = FreeCAD.open(SaveName)
        self.assertTrue(self.Doc.Label_1.Label == "हिन्दी")
        FreeCAD.closeDocument("UnicodeTest")
        FreeCAD.newDocument("SaveRestoreTests")

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("SaveRestoreTests")



#!/usr/bin/env python3
# SPDX-License-Identifier: EUPL-1.2
# A FreeCAD postprocessor targeting Snapmaker machines with CNC capabilities
# ***************************************************************************
# *  Copyright (c) 2025 Clair-Loup Sergent <clsergent@free.fr>              *
# *                                                                         *
# *  Licensed under the EUPL-1.2 with the specific provision                *
# *  (EUPL articles 14 & 15) that the applicable law is the French law.     *
# *  and the Jurisdiction Paris.                                            *
# *  Any redistribution must include the specific provision above.          *
# *                                                                         *
# *  You may obtain a copy of the Licence at:                               *
# *  https://interoperable-europe.ec.europa.eu/collection/eupl/eupl-text-eupl-12 *
# *                                                                         *
# *  Unless required by applicable law or agreed to in writing, software    *
# *  distributed under the Licence is distributed on an "AS IS" basis,      *
# *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or        *
# *  implied. See the Licence for the specific language governing           *
# *  permissions and limitations under the Licence.                         *
# ***************************************************************************


import argparse
import base64
import copy
import datetime
import os
import pathlib
import re
import tempfile
from typing import Any, List, Tuple

import FreeCAD
import Path
import Path.Base.Util as PathUtil
import Path.Post.Processor
import Path.Post.UtilsArguments
import Path.Post.UtilsExport
import Path.Post.Utils as PostUtils
import Path.Post.UtilsParse
import Path.Main.Job

translate = FreeCAD.Qt.translate

if DEBUG := False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


def convert_option_to_attr(option_name):
    # transforms argparse options into identifiers
    if option_name.startswith("--"):
        option_name = option_name[2:]
    elif option_name.startswith("-"):
        option_name = option_name[1:]

    return option_name.replace("-", "_")


SNAPMAKER_MACHINES = dict(
    Original=dict(
        key="Original",
        name="Snapmaker Original",
        boundaries_table=[
            # https://forum.snapmaker.com/t/cnc-work-area-size/5178
            dict(boundaries=dict(X=125, Y=125, Z=50), toolhead="Original_CNC", mods=set()),
        ],
        lead=dict(X=8, Y=8, Z=8),  # Linear module screw pitch (mm/turn)
    ),
    Original_Z_Extension=dict(
        key="Original_Z_Extension",
        name="Snapmaker Original with Z extension",
        boundaries_table=[
            # https://forum.snapmaker.com/t/cnc-work-area-size/5178
            dict(boundaries=dict(X=125, Y=125, Z=146), toolhead="Original_CNC", mods=set()),
        ],
        lead=dict(X=8, Y=8, Z=8),  # Linear module screw pitch (mm/turn)
    ),
    A150=dict(
        key="A150",
        name="Snapmaker 2 A150",
        boundaries_table=[
            # [1] https://support.snapmaker.com/hc/en-us/articles/20786910972311-FAQ-for-Bracing-Kit-for-Snapmaker-2-0-Linear-Modules#h_01HN4Z7S9WJE5BRT492WR0CKH1
            dict(boundaries=dict(X=145, Y=160, Z=90), toolhead="50W_CNC", mods=set()),
            dict(boundaries=dict(X=145, Y=148, Z=90), toolhead="50W_CNC", mods={"BK"}),
        ],
        lead=dict(X=8, Y=8, Z=8),  # Linear module screw pitch (mm/turn)
    ),
    A250=dict(
        key="A250",
        name="Snapmaker 2 A250",
        boundaries_table=[
            # [1] https://support.snapmaker.com/hc/en-us/articles/20786910972311-FAQ-for-Bracing-Kit-for-Snapmaker-2-0-Linear-Modules#h_01HN4Z7S9WJE5BRT492WR0CKH1
            dict(boundaries=dict(X=230, Y=250, Z=180), toolhead="50W_CNC", mods=set()),
            dict(boundaries=dict(X=230, Y=238, Z=180), toolhead="50W_CNC", mods={"BK"}),
            dict(boundaries=dict(X=230, Y=235, Z=180), toolhead="50W_CNC", mods={"QS"}),
            dict(boundaries=dict(X=230, Y=223, Z=180), toolhead="50W_CNC", mods={"BK", "QS"}),
            dict(boundaries=dict(X=230, Y=225, Z=180), toolhead="200W_CNC", mods={"BK"}),
            dict(boundaries=dict(X=230, Y=210, Z=180), toolhead="200W_CNC", mods={"BK", "QS"}),
        ],
        lead=dict(X=8, Y=8, Z=8),  # Linear module screw pitch (mm/turn)
    ),
    A250T=dict(
        key="A250T",
        name="Snapmaker 2 A250T",
        boundaries_table=[
            # [1] https://support.snapmaker.com/hc/en-us/articles/20786910972311-FAQ-for-Bracing-Kit-for-Snapmaker-2-0-Linear-Modules#h_01HN4Z7S9WJE5BRT492WR0CKH1
            dict(boundaries=dict(X=230, Y=250, Z=180), toolhead="50W_CNC", mods=set()),
            dict(boundaries=dict(X=230, Y=238, Z=180), toolhead="50W_CNC", mods={"BK"}),
            dict(boundaries=dict(X=230, Y=235, Z=180), toolhead="50W_CNC", mods={"QS"}),
            dict(boundaries=dict(X=230, Y=223, Z=180), toolhead="50W_CNC", mods={"BK", "QS"}),
            dict(boundaries=dict(X=230, Y=225, Z=180), toolhead="200W_CNC", mods={"BK"}),
            dict(boundaries=dict(X=230, Y=210, Z=180), toolhead="200W_CNC", mods={"BK", "QS"}),
        ],
        lead=dict(X=20, Y=20, Z=8),  # Linear module screw pitch (mm/turn)
    ),
    A350=dict(
        key="A350",
        name="Snapmaker 2 A350",
        boundaries_table=[
            # [1] https://support.snapmaker.com/hc/en-us/articles/20786910972311-FAQ-for-Bracing-Kit-for-Snapmaker-2-0-Linear-Modules#h_01HN4Z7S9WJE5BRT492WR0CKH1
            dict(boundaries=dict(X=320, Y=350, Z=275), toolhead="50W_CNC", mods=set()),
            dict(boundaries=dict(X=320, Y=338, Z=275), toolhead="50W_CNC", mods={"BK"}),
            dict(boundaries=dict(X=320, Y=335, Z=275), toolhead="50W_CNC", mods={"QS"}),
            dict(boundaries=dict(X=320, Y=323, Z=275), toolhead="50W_CNC", mods={"BK", "QS"}),
            dict(boundaries=dict(X=320, Y=325, Z=275), toolhead="200W_CNC", mods={"BK"}),
            dict(boundaries=dict(X=320, Y=310, Z=275), toolhead="200W_CNC", mods={"BK", "QS"}),
        ],
        lead=dict(X=8, Y=8, Z=8),  # Linear module screw pitch (mm/turn)
    ),
    A350T=dict(
        key="A350T",
        name="Snapmaker 2 A350T",
        boundaries_table=[
            # [1] https://support.snapmaker.com/hc/en-us/articles/20786910972311-FAQ-for-Bracing-Kit-for-Snapmaker-2-0-Linear-Modules#h_01HN4Z7S9WJE5BRT492WR0CKH1
            dict(boundaries=dict(X=320, Y=350, Z=275), toolhead="50W_CNC", mods=set()),
            dict(boundaries=dict(X=320, Y=338, Z=275), toolhead="50W_CNC", mods={"BK"}),
            dict(boundaries=dict(X=320, Y=335, Z=275), toolhead="50W_CNC", mods={"QS"}),
            dict(boundaries=dict(X=320, Y=323, Z=275), toolhead="50W_CNC", mods={"BK", "QS"}),
            dict(boundaries=dict(X=320, Y=325, Z=275), toolhead="200W_CNC", mods={"BK"}),
            dict(boundaries=dict(X=320, Y=310, Z=275), toolhead="200W_CNC", mods={"BK", "QS"}),
        ],
        lead=dict(X=20, Y=20, Z=8),  # Linear module screw pitch (mm/turn)
    ),
    Artisan=dict(
        key="Artisan",
        name="Snapmaker Artisan",
        boundaries_table=[
            dict(boundaries=dict(X=400, Y=400, Z=400), toolhead="200W_CNC", mods=set()),
        ],
        lead=dict(X=40, Y=40, Z=8),  # Linear module screw pitch (mm/turn)
    ),
)

# These modifications were released to upgrade the Snapmaker 2.0 machines
# which started on Kickstarter.
SNAPMAKER_MOD_KITS = {
    "QS": dict(
        key="QS",
        name="Quick Swap Kit",
        option_name="--quick-swap",
        option_help_text="Indicates that the quick swap kit is installed. Only compatible with Snapmaker 2 machines.",
    ),
    "BK": dict(
        key="BK",
        name="Bracing Kit",
        option_name="--bracing-kit",
        option_help_text="Indicates that the bracing kit is installed. Only compatible with Snapmaker 2 machines.",
    ),
}

# Could support other types of toolheads (laser, drag knife, 3DP, ...) in the future
# https://wiki.snapmaker.com/en/Snapmaker_Luban/manual/2_supported_gcode_references#m3m4-modified-cnclaser-on
SNAPMAKER_TOOLHEADS = {
    "Original_CNC": dict(
        key="Original_CNC",
        name="Original CNC module",
        speed_rpm=dict(min=0, max=7000),
        boundaries_delta=dict(X=0, Y=0, Z=0),
        has_percent=True,
        has_speed_s=False,
    ),
    "50W_CNC": dict(
        key="50W_CNC",
        name="50W CNC module",
        speed_rpm=dict(min=0, max=12000),
        boundaries_delta=dict(X=0, Y=0, Z=0),
        has_percent=True,
        has_speed_s=False,
    ),
    "200W_CNC": dict(
        key="200W_CNC",
        name="200W CNC module",
        speed_rpm=dict(min=0, max=18000),
        boundaries_delta=dict(X=0, Y=-13, Z=0),
        has_percent=True,
        has_speed_s=True,
    ),
}


class CoordinatesAction(argparse.Action):
    """argparse Action to handle coordinates (x,y,z)"""

    def __call__(self, parser, namespace, values, option_string=None):
        match = re.match(
            r"^\s*(?P<X>-?\d+\.?\d*),?\s*(?P<Y>-?\d+\.?\d*),?\s*(?P<Z>-?\d+\.?\d*)\s*$", values
        )
        if match:
            # setattr(namespace, self.dest, 'G0 X{0} Y{1} Z{2}'.format(*match.groups()))
            params = {key: float(value) for key, value in match.groupdict().items()}
            setattr(namespace, self.dest, params)
        else:
            raise argparse.ArgumentError(None, message="invalid coordinates provided")


class ExtremaAction(argparse.Action):
    """argparse Action to handle integer extrema (min,max)"""

    def __call__(self, parser, namespace, values, option_string=None):
        if match := re.match(r"^ *(\d+),? *(\d+) *$", values):
            # setattr(namespace, self.dest, 'G0 X{0} Y{1} Z{2}'.format(*match.groups()))
            params = {
                key: int(value)
                for key, value in zip(
                    (
                        "min",
                        "max",
                    ),
                    match.groups(),
                )
            }
            setattr(namespace, self.dest, params)
        else:
            raise argparse.ArgumentError(None, message="invalid values provided, should be int,int")


class Snapmaker(Path.Post.Processor.PostProcessor):
    """FreeCAD postprocessor targeting Snapmaker machines with CNC capabilities"""

    def __init__(self, job) -> None:
        super().__init__(
            job=job,
            tooltip=translate("CAM", "Snapmaker post processor"),
            tooltipargs=[""],
            units="Metric",
        )

        self.initialize()

    def initialize(self):
        """initialize values and arguments"""
        self.values: dict[str, Any] = dict()
        self.argument_defaults: dict[str, bool] = dict()
        self.arguments_visible: dict[str, bool] = dict()
        self.parser = argparse.ArgumentParser()

        self.snapmaker_init_values()
        self.snapmaker_init_argument_defaults()
        self.snapmaker_init_arguments_visible()
        self.parser = self.snapmaker_init_parser(
            self.values, self.argument_defaults, self.arguments_visible
        )

        # create another parser with all visible arguments
        all_arguments_visible = dict()
        for key in iter(self.arguments_visible):
            all_arguments_visible[key] = True
        self.visible_parser = self.snapmaker_init_parser(
            self.values, self.argument_defaults, all_arguments_visible
        )

        FreeCAD.Console.PrintLog(f'{self.values["POSTPROCESSOR_FILE_NAME"]}: initialized.\n')

    def snapmaker_init_values(self):
        """Initialize values that are used throughout the postprocessor."""
        Path.Post.UtilsArguments.init_shared_values(self.values)

        # shared values
        self.values["POSTPROCESSOR_FILE_NAME"] = __name__
        self.values["COMMENT_SYMBOL"] = ";"
        self.values["ENABLE_MACHINE_SPECIFIC_COMMANDS"] = True
        self.values["END_OF_LINE_CHARACTERS"] = "\n"
        self.values["FINISH_LABEL"] = "End"
        self.values["LINE_INCREMENT"] = 1
        self.values["MACHINE_NAME"] = None
        self.values["MODAL"] = False
        self.values["OUTPUT_PATH_LABELS"] = True
        self.values["OUTPUT_HEADER"] = (
            True  # remove FreeCAD standard header and use a custom Snapmaker Header
        )
        self.values["OUTPUT_TOOL_CHANGE"] = True
        self.values["PARAMETER_ORDER"] = [
            "X",
            "Y",
            "Z",
            "A",
            "B",
            "C",
            "I",
            "J",
            "F",
            "S",
            "T",
            "Q",
            "R",
            "L",
            "H",
            "D",
            "P",
            "O",
        ]
        self.values["PREAMBLE"] = """G90\nG17"""
        self.values["PRE_OPERATION"] = """"""
        self.values["POST_OPERATION"] = """"""
        self.values["POSTAMBLE"] = """M400\nM5"""
        self.values["SHOW_MACHINE_UNITS"] = False
        self.values["SPINDLE_DECIMALS"] = 0
        self.values["SPINDLE_WAIT"] = 4.0
        self.values["TOOL_CHANGE"] = "M76"  # handle tool change by inserting an HMI pause
        self.values["TRANSLATE_DRILL_CYCLES"] = True  # drill cycle gcode must be translated
        self.values["USE_TLO"] = False  # G43 is not handled.

        # snapmaker values
        self.values["THUMBNAIL"] = True
        self.values["BOUNDARIES"] = None
        self.values["BOUNDARIES_CHECK"] = False
        self.values["MACHINES"] = SNAPMAKER_MACHINES
        self.values["MOD_KITS_ALL"] = SNAPMAKER_MOD_KITS
        self.values["TOOLHEADS"] = SNAPMAKER_TOOLHEADS
        self.values["TOOLHEAD_NAME"] = None
        self.values["SPINDLE_SPEEDS"] = dict()
        self.values["SPINDLE_PERCENT"] = None

    def snapmaker_init_argument_defaults(self) -> None:
        """Initialize which arguments (in a pair) are shown as the default argument."""
        Path.Post.UtilsArguments.init_argument_defaults(self.argument_defaults)

        self.argument_defaults["tlo"] = False
        self.argument_defaults["translate-drill"] = True

        # snapmaker arguments
        self.argument_defaults["thumbnail"] = True
        self.argument_defaults["gui"] = True
        self.argument_defaults["boundaries-check"] = True

    def snapmaker_init_arguments_visible(self) -> None:
        """Initialize which argument pairs are visible in TOOLTIP_ARGS."""
        Path.Post.UtilsArguments.init_arguments_visible(self.arguments_visible)

        self.arguments_visible["axis-modal"] = False
        self.arguments_visible["header"] = False
        self.arguments_visible["return-to"] = True
        self.arguments_visible["tlo"] = False
        self.arguments_visible["tool_change"] = True
        self.arguments_visible["translate-drill"] = False
        self.arguments_visible["wait-for-spindle"] = True

        # snapmaker arguments (for record, always visible)
        self.arguments_visible["thumbnail"] = True
        self.arguments_visible["gui"] = True
        self.arguments_visible["boundaries"] = True
        self.arguments_visible["boundaries-check"] = True
        self.arguments_visible["machine"] = True
        self.arguments_visible["toolhead"] = True
        self.arguments_visible["line-increment"] = True
        self.arguments_visible["spindle-speeds"] = True

    def snapmaker_init_parser(
        self, values, argument_defaults, arguments_visible
    ) -> argparse.ArgumentParser:
        """Initialize the postprocessor arguments parser"""
        parser = Path.Post.UtilsArguments.init_shared_arguments(
            values, argument_defaults, arguments_visible
        )

        # snapmaker custom arguments
        group = parser.add_argument_group("Snapmaker only arguments")
        # add_flag_type_arguments function is not used as its behavior is inconsistent with argparse
        # handle thumbnail generation
        group.add_argument(
            "--thumbnail",
            action="store_true",
            default=argument_defaults["thumbnail"],
            help="Include a thumbnail (require --gui)",
        )
        group.add_argument(
            "--no-thumbnail", action="store_false", dest="thumbnail", help="Remove thumbnail"
        )

        group.add_argument(
            "--gui",
            action="store_true",
            default=argument_defaults["gui"],
            help="allow the postprocessor to execute GUI methods",
        )
        group.add_argument(
            "--no-gui",
            action="store_false",
            dest="gui",
            help="Execute postprocessor without requiring GUI",
        )

        group.add_argument(
            "--boundaries-check",
            action="store_true",
            default=argument_defaults["boundaries-check"],
            help="check boundaries according to the machine build area",
        )
        group.add_argument(
            "--no-boundaries-check",
            action="store_false",
            dest="boundaries_check",
            help="Disable boundaries check",
        )

        group.add_argument(
            "--boundaries",
            action=CoordinatesAction,
            default=None,
            help='Custom boundaries (e.g. "100, 200, 300"). Overrides boundaries from --machine',
        )

        group.add_argument(
            "--machine",
            default=None,
            required=True,
            choices=self.values["MACHINES"].keys(),
            help=f"Snapmaker machine. Choose from [{self.values['MACHINES'].keys()}].",
        )

        for key, value in SNAPMAKER_MOD_KITS.items():
            group.add_argument(
                value["option_name"],
                default=False,
                action="store_true",
                help=value["option_help_text"],
            )

        group.add_argument(
            "--toolhead",
            default=None,
            choices=self.values["TOOLHEADS"].keys(),
            help=f"Snapmaker toolhead. Choose from [{self.values['TOOLHEADS'].keys()}].",
        )

        group.add_argument(
            "--spindle-speeds",
            action=ExtremaAction,
            default=None,
            help="Set minimum/maximum spindle speeds as --spindle-speeds='min,max'",
        )

        group.add_argument(
            "--spindle-percent",
            action="store_true",
            default=None,
            help="use percent as toolhead spindle speed unit (default: use RPM if supported by toolhead, otherwise percent)",
        )

        group.add_argument(
            "--line-number",
            type=int,
            default=self.values["line_number"],
            help="Set the line starting value",
        )

        group.add_argument(
            "--line-increment",
            type=int,
            default=self.values["LINE_INCREMENT"],
            help="Set the line increment value",
        )

        return parser

    def snapmaker_process_arguments(
        self, filename: str = "-"
    ) -> Tuple[bool, str | argparse.Namespace]:
        """Process any arguments to the postprocessor."""
        (flag, args) = Path.Post.UtilsArguments.process_shared_arguments(
            self.values, self.parser, self._job.PostProcessorArgs, self.visible_parser, filename
        )
        if flag:  # process extra arguments only if flag is True
            self._units = self.values["UNITS"]

            # --machine is a required "option"
            machine = self.values["MACHINES"][args.machine]
            self.values["MACHINE_KEY"] = machine["key"]
            self.values["MACHINE_NAME"] = machine["name"]

            compatible_toolheads = {bt["toolhead"] for bt in machine["boundaries_table"]}

            if args.toolhead:
                if args.toolhead not in compatible_toolheads:
                    FreeCAD.Console.PrintError(
                        f"Selected --toolhead={args.toolhead} is not compatible with machine {machine['name']}."
                        + f" Choose from [{compatible_toolheads}]\n"
                    )
                    flag = False
                    return (flag, args)
                toolhead = self.values["TOOLHEADS"][args.toolhead]
            elif len(compatible_toolheads) == 1:
                toolhead_key = next(iter(compatible_toolheads))
                toolhead = self.values["TOOLHEADS"][toolhead_key]
            else:
                FreeCAD.Console.PrintError(
                    f"Machine {machine['name']} has multiple compatible toolheads:\n"
                    f"{compatible_toolheads}\n"
                    "Please add --toolhead argument.\n"
                )
                flag = False
                return (flag, args)
            self.values["TOOLHEAD_KEY"] = toolhead["key"]
            self.values["TOOLHEAD_NAME"] = toolhead["name"]

            self.values["SPINDLE_SPEEDS"] = toolhead["speed_rpm"]

            if args.spindle_speeds:  # may override toolhead value, which is expected
                self.values["SPINDLE_SPEEDS"] = args.spindle_speeds

            if args.spindle_percent is not None:
                if toolhead["has_percent"]:
                    self.values["SPINDLE_PERCENT"] = True
                else:
                    FreeCAD.Console.PrintError(
                        f"Requested spindle speed in percent, but toolhead {toolhead['name']}"
                        + " does not support speed as percent.\n"
                    )
                    flag = False
                    return (flag, args)
            else:
                # Prefer speed S over percent P
                self.values["SPINDLE_PERCENT"] = (
                    toolhead["has_percent"] and not toolhead["has_speed_s"]
                )
            if self.values["SPINDLE_PERCENT"]:
                FreeCAD.Console.PrintWarning(
                    "Spindle speed will be controlled using using percentages.\n"
                )
            else:
                FreeCAD.Console.PrintWarning("Spindle speed will be controlled using using RPM.\n")

            self.values["MOD_KITS_INSTALLED"] = []
            if args.boundaries:  # may override machine boundaries, which is expected
                self.values["BOUNDARIES"] = args.boundaries
                self.values["MACHINE_NAME"] += " Boundaries override=" + str(args.boundaries)
            else:
                compatible_modkit_combos = [
                    bt["mods"]
                    for bt in machine["boundaries_table"]
                    if toolhead["key"] == bt["toolhead"]
                ]
                configured_modkits = set()

                # Determine which mod kits are requested from the options
                for mod_kit in self.values["MOD_KITS_ALL"].values():
                    if getattr(args, convert_option_to_attr(mod_kit["option_name"])):
                        configured_modkits.add(mod_kit["key"])
                        self.values["MACHINE_NAME"] += " " + mod_kit["name"]
                        self.values["MOD_KITS_INSTALLED"].append(mod_kit["key"])

                if configured_modkits not in compatible_modkit_combos:
                    FreeCAD.Console.PrintError(
                        f"Machine {machine['name']} with toolhead {toolhead['name']}"
                        + f" is not compatible with modkit {configured_modkits if configured_modkits else None}.\n"
                        + f" Choose from {compatible_modkit_combos}."
                    )
                    flag = False
                    return (flag, args)

                # Update machine dimensions based on installed toolhead and mod kits
                boundaries_table_entry_l = [
                    bt
                    for bt in machine["boundaries_table"]
                    if bt["toolhead"] == toolhead["key"] and bt["mods"] == configured_modkits
                ]
                assert len(boundaries_table_entry_l) == 1
                boundaries_table_entry = boundaries_table_entry_l[0]

                # The deepcopy is necessary to avoid modifying the boundaries in the MACHINES dict.
                self.values["BOUNDARIES"] = copy.deepcopy(boundaries_table_entry["boundaries"])
                self.values["MACHINE_NAME"] += " " + toolhead["name"]

            self.values["THUMBNAIL"] = args.thumbnail
            self.values["ALLOW_GUI"] = args.gui
            self.values["line_number"] = args.line_number
            self.values["LINE_INCREMENT"] = args.line_increment

            if args.boundaries_check and not self.values["BOUNDARIES"]:
                FreeCAD.Console.PrintError("Boundary check skipped: no valid boundaries supplied\n")
                self.values["BOUNDARIES_CHECK"] = False
            else:
                self.values["BOUNDARIES_CHECK"] = args.boundaries_check

        return flag, args

    def snapmaker_process_postables(self, filename: str = "-") -> List[Tuple[str, str]]:
        """process job sections to gcode"""
        sections: List[Tuple[str, str]] = list()

        postables = self._buildPostList()

        # basic filename handling
        if len(postables) > 1 and filename != "-":
            filename = pathlib.Path(filename)
            filename = str(filename.with_stem(filename.stem + "_{name}"))

        for name, objects in postables:
            gcode = self.export_common(objects, filename.format(name=name))
            sections.append((name, gcode))

        return sections

    def get_thumbnail(self) -> str:
        """generate a thumbnail of the job from the given objects"""
        if self.values["THUMBNAIL"] is False:
            return "thumbnail: deactivated."

        if not (self.values["ALLOW_GUI"] and FreeCAD.GuiUp):
            FreeCAD.Console.PrintError(
                "GUI access required: thumbnail generation skipped. Consider adding --gui\n"
            )
            return "thumbnail: GUI required."

        # get FreeCAD references
        import FreeCADGui

        view = FreeCADGui.activeDocument().activeView()
        selection = FreeCADGui.Selection

        # save current selection
        selected = [
            obj.Object for obj in selection.getCompleteSelection() if hasattr(obj, "Object")
        ]
        selection.clearSelection()

        # clear view
        FreeCADGui.runCommand("Std_SelectAll", 0)
        all = []
        for obj in selection.getCompleteSelection():
            if hasattr(obj, "Object"):
                all.append((obj.Object, obj.Object.Visibility))
                obj.Object.ViewObject.hide()

        # select models to display
        for model in self._job.Model.Group:
            model.ViewObject.show()
            selection.addSelection(model.Document.Name, model.Name)
        view.fitAll()  # center selection
        view.viewIsometric()  # display as isometric
        selection.clearSelection()

        # generate thumbnail
        with tempfile.TemporaryDirectory() as temp:
            path = os.path.join(temp, "thumbnail.png")
            view.saveImage(path, 720, 480, "Transparent")
            with open(path, "rb") as file:
                data = file.read()

        # restore view
        for obj, visibility in all:
            if visibility:
                obj.ViewObject.show()

        # restore selection
        for obj in selected:
            selection.clearSelection()
            selection.addSelection(obj.Document.Name, obj.Name)

        return f"thumbnail: data:image/png;base64,{base64.b64encode(data).decode()}"

    def output_header(self, gcode: List[str]):
        """custom method derived from Path.Post.UtilsExport.output_header"""
        cam_file: str
        comment: str

        if not self.values["OUTPUT_HEADER"]:
            return

        def add_comment(text):
            com = Path.Post.UtilsParse.create_comment(self.values, text)
            gcode.append(f"{Path.Post.UtilsParse.linenumber(self.values)}{com}")

        add_comment("Header Start")
        add_comment("header_type: cnc")
        add_comment(f'machine: {self.values["MACHINE_NAME"]}')
        comment = Path.Post.UtilsParse.create_comment(
            self.values, f'Post Processor: {self.values["POSTPROCESSOR_FILE_NAME"]}'
        )
        gcode.append(f"{Path.Post.UtilsParse.linenumber(self.values)}{comment}")
        if FreeCAD.ActiveDocument:
            cam_file = os.path.basename(FreeCAD.ActiveDocument.FileName)
        else:
            cam_file = "<None>"
        add_comment(f"CAM File: {cam_file}")
        add_comment(f"Output Time: {datetime.datetime.now()}")
        add_comment(self.get_thumbnail())

    def convert_spindle(self, gcode: List[str]) -> List[str]:
        """convert spindle speed values from RPM to percent (%) (M3/M4 commands)"""
        if self.values["SPINDLE_PERCENT"] is False:
            return gcode

        # https://wiki.snapmaker.com/en/Snapmaker_Luban/manual/2_supported_gcode_references#m3m4-modified-cnclaser-on
        # Speed as percentage in [0,100]% range
        for index, commandline in enumerate(
            gcode
        ):  # .split(self.values["END_OF_LINE_CHARACTERS"]):
            if match := re.match(r"(?P<command>M0?[34])\D.*(?P<spindle>S\d+.?\d*)", commandline):
                percent = (
                    float(match.group("spindle")[1:]) * 100 / self.values["SPINDLE_SPEEDS"]["max"]
                )
                gcode[index] = (
                    gcode[index][: match.span("spindle")[0]]
                    + f'P{percent:.{self.values["SPINDLE_DECIMALS"]}f}'
                    + gcode[index][match.span("spindle")[1] :]
                )
        return gcode

    def check_boundaries(self, gcode: List[str]) -> bool:
        """Check boundaries and return whether it succeeded"""
        status = True
        FreeCAD.Console.PrintLog("Boundaries check\n")

        extrema = dict(X=[0, 0], Y=[0, 0], Z=[0, 0])
        position = dict(X=0, Y=0, Z=0)
        relative = False

        for index, commandline in enumerate(gcode):
            if re.match(r"G90(?:\D|$)", commandline):
                relative = False
            elif re.match(r"G91(?:\D|$)", commandline):
                relative = True
            elif re.match(r"G0?[12](?:\D|$)", commandline):
                for axis, value in re.findall(
                    r"(?P<axis>[XYZ])(?P<value>-?\d+\.?\d*)(?:\D|$)", commandline
                ):
                    if relative:
                        position[axis] += float(value)
                    else:
                        position[axis] = float(value)
                    extrema[axis][0] = min(extrema[axis][0], position[axis])
                    extrema[axis][1] = max(extrema[axis][1], position[axis])

        for axis in extrema.keys():
            if abs(extrema[axis][1] - extrema[axis][0]) > self.values["BOUNDARIES"][axis]:
                # gcode.insert(0, f';WARNING: Boundary check: job exceeds machine limit on {axis} axis{self.values["END_OF_LINE_CHARACTERS"]}')
                FreeCAD.Console.PrintWarning(
                    f"Boundary check: job exceeds machine limit on {axis} axis\n"
                )
                status = False

        return status

    def export_common(self, objects: List, filename: str | pathlib.Path) -> str:
        """custom method derived from Path.Post.UtilsExport.export_common"""
        final: str
        gcode: List = []
        result: bool

        for obj in objects:
            if not hasattr(obj, "Path"):
                print(f"The object {obj.Name} is not a path.")
                print("Please select only path and Compounds.")
                return ""

        Path.Post.UtilsExport.check_canned_cycles(self.values)
        self.output_header(gcode)
        Path.Post.UtilsExport.output_safetyblock(self.values, gcode)
        Path.Post.UtilsExport.output_tool_list(self.values, gcode, objects)
        Path.Post.UtilsExport.output_preamble(self.values, gcode)
        Path.Post.UtilsExport.output_motion_mode(self.values, gcode)
        Path.Post.UtilsExport.output_units(self.values, gcode)

        for obj in objects:
            # Skip inactive operations
            if not PathUtil.activeForOp(obj):
                continue
            coolant_mode = PathUtil.coolantModeForOp(obj)
            Path.Post.UtilsExport.output_start_bcnc(self.values, gcode, obj)
            Path.Post.UtilsExport.output_preop(self.values, gcode, obj)
            Path.Post.UtilsExport.output_coolant_on(self.values, gcode, coolant_mode)
            # output the G-code for the group (compound) or simple path
            Path.Post.UtilsParse.parse_a_group(self.values, gcode, obj)

            Path.Post.UtilsExport.output_postop(self.values, gcode, obj)
            Path.Post.UtilsExport.output_coolant_off(self.values, gcode, coolant_mode)

        Path.Post.UtilsExport.output_return_to(self.values, gcode)
        #
        # This doesn't make sense to me.  It seems that both output_start_bcnc and
        # output_end_bcnc should be in the for loop or both should be out of the
        # for loop.  However, that is the way that grbl post code was written, so
        # for now I will leave it that way until someone has time to figure it out.
        #
        Path.Post.UtilsExport.output_end_bcnc(self.values, gcode)
        Path.Post.UtilsExport.output_postamble_header(self.values, gcode)
        Path.Post.UtilsExport.output_tool_return(self.values, gcode)
        Path.Post.UtilsExport.output_safetyblock(self.values, gcode)
        Path.Post.UtilsExport.output_postamble(self.values, gcode)
        gcode = self.convert_spindle(gcode)

        if self.values["BOUNDARIES_CHECK"]:
            self.check_boundaries(gcode)

        # add the appropriate end-of-line characters to the gcode, including after the last line
        gcode.append("")
        final = self.values["END_OF_LINE_CHARACTERS"].join(gcode)

        if FreeCAD.GuiUp and self.values["SHOW_EDITOR"]:
            # size limit removed as irrelevant on my computer - see if issues occur
            dia = PostUtils.GCodeEditorDialog()
            # the editor expects lines to end in "\n", and returns lines ending in "\n"
            if self.values["END_OF_LINE_CHARACTERS"] == "\n":
                dia.editor.setText(final)
                if dia.exec_():
                    final = dia.editor.toPlainText()
            else:
                final_for_editor = "\n".join(gcode)
                dia.editor.setText(final_for_editor)
                if dia.exec_():
                    final_for_editor = dia.editor.toPlainText()
                    # convert all "\n" to the appropriate end-of-line characters
                    final = final_for_editor.replace("\n", self.values["END_OF_LINE_CHARACTERS"])

        if not filename == "-":
            with open(filename, "w", encoding="utf-8", newline="") as gfile:
                gfile.write(final)

        return final

    def export(self, filename: str | pathlib.Path = "-"):
        """process gcode and export"""
        (flag, args) = self.snapmaker_process_arguments()
        if flag:
            return self.snapmaker_process_postables(filename)
        if args is None:
            return None
        return [("allitems", args)]

    @property
    def tooltip(self) -> str:
        tooltip = "Postprocessor of the FreeCAD CAM workbench for the Snapmaker machines"
        return tooltip

    @property
    def tooltipArgs(self) -> str:
        return self.parser.format_help()


if __name__ == "__main__":
    Snapmaker(None).visible_parser.format_help()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import Path
import Path.Base.Util as PathUtil
import Path.Dressup.Utils as PathDressup
import Path.Main.Stock as PathStock
import PathScripts.PathUtils as PathUtils

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


def _vstr(v):
    if v:
        return "(%.2f, %.2f, %.2f)" % (v.x, v.y, v.z)
    return "-"


class DressupPathBoundary(object):
    def __init__(self, obj, base, job):
        obj.addProperty(
            "App::PropertyLink",
            "Base",
            "Base",
            QT_TRANSLATE_NOOP("App::Property", "The base path to modify"),
        )
        obj.Base = base
        obj.addProperty(
            "App::PropertyLink",
            "Stock",
            "Boundary",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Solid object to be used to limit the generated Path.",
            ),
        )
        obj.Stock = PathStock.CreateFromBase(job)
        obj.addProperty(
            "App::PropertyBool",
            "Inside",
            "Boundary",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Determines if Boundary describes an inclusion or exclusion mask.",
            ),
        )
        obj.Inside = True
        obj.addProperty(
            "App::PropertyBool",
            "KeepToolDown",
            "Boundary",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Keep tool down.",
            ),
        )
        obj.KeepToolDown = False

        self.obj = obj
        self.safeHeight = None
        self.clearanceHeight = None

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onChanged(self, obj, prop):
        if prop == "Path" and obj.ViewObject:
            obj.ViewObject.signalChangeIcon()

    def onDocumentRestored(self, obj):
        self.obj = obj
        if not hasattr(obj, "KeepToolDown"):
            obj.addProperty(
                "App::PropertyBool",
                "KeepToolDown",
                "Boundary",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Keep tool down.",
                ),
            )

    def onDelete(self, obj, args):
        if obj.Base:
            job = PathUtils.findParentJob(obj)
            if job:
                job.Proxy.addOperation(obj.Base, obj)
            if obj.Base.ViewObject:
                obj.Base.ViewObject.Visibility = True
            obj.Base = None
        if obj.Stock:
            obj.Document.removeObject(obj.Stock.Name)
            obj.Stock = None
        return True

    def execute(self, obj):
        pb = PathBoundary(obj.Base, obj.Stock.Shape, obj.Inside, obj.KeepToolDown)
        obj.Path = pb.execute()


# Eclass


class PathBoundary:
    """class PathBoundary...
    This class requires a base operation, boundary shape, and optional inside boolean (default is True).
    The `execute()` method returns a Path object with path commands limited to cut paths inside or outside
    the provided boundary shape.
    """

    def __init__(self, baseOp, boundaryShape, inside=True, keepToolDown=False):
        self.baseOp = baseOp
        self.boundary = boundaryShape
        self.inside = inside
        self.safeHeight = None
        self.clearanceHeight = None
        self.strG0ZsafeHeight = None
        self.strG0ZclearanceHeight = None
        self.keepToolDown = keepToolDown

    def boundaryCommands(
        self, begin, end, verticalFeed, horizFeed=None, keepToolDown=False, isStartMovements=False
    ):
        Path.Log.track(_vstr(begin), _vstr(end))
        if end and Path.Geom.pointsCoincide(begin, end):
            return []
        cmds = []

        if isStartMovements or not keepToolDown:
            if begin.z < self.safeHeight:
                cmds.append(self.strG0ZsafeHeight)
            if begin.z < self.clearanceHeight:
                cmds.append(self.strG0ZclearanceHeight)
            if end:
                cmds.append(Path.Command("G0", {"X": end.x, "Y": end.y}))
                if end.z < self.clearanceHeight:
                    cmds.append(Path.Command("G0", {"Z": max(self.safeHeight, end.z)}))
                if end.z < self.safeHeight:
                    cmds.append(Path.Command("G1", {"Z": end.z, "F": verticalFeed}))
        else:
            if end:
                if horizFeed and Path.Geom.isRoughly(begin.z, end.z, 0.001):
                    speed = horizFeed
                else:
                    speed = verticalFeed
                cmds.append(Path.Command("G1", {"X": end.x, "Y": end.y, "Z": end.z, "F": speed}))

        return cmds

    def execute(self):
        if (
            not self.baseOp
            or not self.baseOp.isDerivedFrom("Path::Feature")
            or not self.baseOp.Path
        ):
            return None

        path = PathUtils.getPathWithPlacement(self.baseOp)
        if len(path.Commands) == 0:
            Path.Log.warning("No Path Commands for %s" % self.baseOp.Label)
            return []

        tc = PathDressup.toolController(self.baseOp)

        self.safeHeight = float(PathUtil.opProperty(self.baseOp, "SafeHeight"))
        self.clearanceHeight = float(PathUtil.opProperty(self.baseOp, "ClearanceHeight"))
        self.strG0ZsafeHeight = Path.Command(  # was a Feed rate with G1
            "G0", {"Z": self.safeHeight, "F": tc.VertRapid.Value}
        )
        self.strG0ZclearanceHeight = Path.Command("G0", {"Z": self.clearanceHeight})

        cmd = path.Commands[0]
        pos = cmd.Placement.Base  # bogus m/c position to create first edge
        bogusX = True
        bogusY = True
        commands = [cmd]
        lastExit = None
        isStartMovements = True
        for cmd in path.Commands[1:]:
            if cmd.Name in Path.Geom.CmdMoveAll:
                if bogusX:
                    bogusX = "X" not in cmd.Parameters
                if bogusY:
                    bogusY = "Y" not in cmd.Parameters
                edge = Path.Geom.edgeForCmd(cmd, pos)
                if edge and cmd.Name in Path.Geom.CmdMoveDrill:
                    inside = edge.common(self.boundary).Edges
                    outside = edge.cut(self.boundary).Edges
                    if 1 == len(inside) and 0 == len(outside):
                        commands.append(cmd)
                if edge and not cmd.Name in Path.Geom.CmdMoveDrill:
                    inside = edge.common(self.boundary).Edges
                    outside = edge.cut(self.boundary).Edges
                    if not self.inside:  # UI "inside boundary" param
                        tmp = inside
                        inside = outside
                        outside = tmp
                    # it's really a shame that one cannot trust the sequence and/or
                    # orientation of edges
                    if 1 == len(inside) and 0 == len(outside):
                        Path.Log.track(_vstr(pos), _vstr(lastExit), " + ", cmd)
                        # cmd fully included by boundary
                        if lastExit:
                            if not (
                                bogusX or bogusY
                            ):  # don't insert false paths based on bogus m/c position
                                commands.extend(
                                    self.boundaryCommands(lastExit, pos, tc.VertFeed.Value)
                                )
                            lastExit = None
                        commands.append(cmd)
                        pos = Path.Geom.commandEndPoint(cmd, pos)
                    elif 0 == len(inside) and 1 == len(outside):
                        Path.Log.track(_vstr(pos), _vstr(lastExit), " - ", cmd)
                        # cmd fully excluded by boundary
                        if not lastExit:
                            lastExit = pos
                        pos = Path.Geom.commandEndPoint(cmd, pos)
                    else:
                        Path.Log.track(_vstr(pos), _vstr(lastExit), len(inside), len(outside), cmd)
                        # cmd pierces boundary
                        while inside or outside:
                            ie = [e for e in inside if Path.Geom.edgeConnectsTo(e, pos)]
                            Path.Log.track(ie)
                            if ie:
                                e = ie[0]
                                LastPt = e.valueAt(e.LastParameter)
                                flip = Path.Geom.pointsCoincide(pos, LastPt)
                                newPos = e.valueAt(e.FirstParameter) if flip else LastPt
                                # inside edges are taken at this point (see swap of inside/outside
                                # above - so we can just connect the dots ...
                                if lastExit:
                                    if not (bogusX or bogusY):
                                        commands.extend(
                                            self.boundaryCommands(
                                                lastExit,
                                                pos,
                                                tc.VertFeed.Value,
                                                tc.HorizFeed.Value,
                                                self.keepToolDown,
                                                isStartMovements,
                                            )
                                        )
                                        isStartMovements = False
                                    lastExit = None
                                Path.Log.track(e, flip)
                                if not (
                                    bogusX or bogusY
                                ):  # don't insert false paths based on bogus m/c position
                                    commands.extend(
                                        Path.Geom.cmdsForEdge(
                                            e,
                                            flip,
                                            False,
                                            50,
                                            tc.HorizFeed.Value,
                                            tc.VertFeed.Value,
                                        )
                                    )
                                inside.remove(e)
                                pos = newPos
                                lastExit = newPos
                            else:
                                oe = [e for e in outside if Path.Geom.edgeConnectsTo(e, pos)]
                                Path.Log.track(oe)
                                if oe:
                                    e = oe[0]
                                    ptL = e.valueAt(e.LastParameter)
                                    flip = Path.Geom.pointsCoincide(pos, ptL)
                                    newPos = e.valueAt(e.FirstParameter) if flip else ptL
                                    # outside edges are never taken at this point (see swap of
                                    # inside/outside above) - so just move along ...
                                    outside.remove(e)
                                    pos = newPos
                                else:
                                    Path.Log.error("huh?")
                                    import Part

                                    Part.show(Part.Vertex(pos), "pos")
                                    for e in inside:
                                        Part.show(e, "ei")
                                    for e in outside:
                                        Part.show(e, "eo")
                                    raise Exception("This is not supposed to happen")
                                # Eif
                            # Eif
                        # Ewhile
                    # Eif
                    # pos = Path.Geom.commandEndPoint(cmd, pos)
                # Eif
            else:
                Path.Log.track("no-move", cmd)
                commands.append(cmd)
        if lastExit:
            commands.extend(self.boundaryCommands(lastExit, None, tc.VertFeed.Value))
            lastExit = None

        Path.Log.track(commands)
        return Path.Path(commands)


# Eclass


def Create(base, name="DressupPathBoundary"):
    """Create(base, name='DressupPathBoundary') ... creates a dressup limiting base's Path to a boundary."""

    if not base.isDerivedFrom("Path::Feature"):
        Path.Log.error(
            translate("CAM_DressupPathBoundary", "The selected object is not a path") + "\n"
        )
        return None

    obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    job = PathUtils.findParentJob(base)
    obj.Proxy = DressupPathBoundary(obj, base, job)
    job.Proxy.addOperation(obj, base, True)
    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 Shai Seger <shaise at gmail>                       *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""
Command and task window handler for the OpenGL based CAM simulator
"""


import math
import os
import FreeCAD
import Path.Base.Util as PathUtil
import Path.Dressup.Utils as PathDressup
import Path.Main.Job as PathJob
from PathScripts import PathUtils
import CAMSimulator

from FreeCAD import Vector, Placement, Rotation


# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Mesh = LazyLoader("Mesh", globals(), "Mesh")
Part = LazyLoader("Part", globals(), "Part")

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtGui, QtCore
    from PySide.QtGui import QDialogButtonBox

_filePath = os.path.dirname(os.path.abspath(__file__))


def IsSame(x, y):
    """Check if two floats are the same within an epsilon"""
    return abs(x - y) < 0.0001


def RadiusAt(edge, p):
    """Find the tool radius within a point on its circumference"""
    x = edge.valueAt(p).x
    y = edge.valueAt(p).y
    return math.sqrt(x * x + y * y)


class CAMSimTaskUi:
    """Handles the simulator task panel"""

    def __init__(self, parent):
        # this will create a Qt widget from our ui file
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/TaskCAMSimulator.ui")
        self.parent = parent

    def getStandardButtons(self, *_args):
        """Task panel needs only Close button"""
        return QDialogButtonBox.Close

    def reject(self):
        """User Pressed the Close button"""
        self.parent.cancel()
        FreeCADGui.Control.closeDialog()


def TSError(msg):
    """Display error message"""
    QtGui.QMessageBox.information(None, "Path Simulation", msg)


class CAMSimulation:
    """Handles and prepares CAM jobs for simulation"""

    def __init__(self):
        self.debug = False
        self.stdrot = FreeCAD.Rotation(Vector(0, 0, 1), 0)
        self.iprogress = 0
        self.numCommands = 0
        self.simperiod = 20
        self.quality = 10
        self.resetSimulation = False
        self.jobs = []
        self.initdone = False
        self.taskForm = None
        self.disableAnim = False
        self.firstDrill = True
        self.millSim = None
        self.job = None
        self.activeOps = []
        self.ioperation = 0
        self.stock = None
        self.busy = False
        self.operations = []
        self.baseShape = None

    def Connect(self, but, sig):
        """Connect task panel buttons"""
        QtCore.QObject.connect(but, QtCore.SIGNAL("clicked()"), sig)

    def FindClosestEdge(self, edges, px, pz):
        """Convert tool shape to tool profile needed by GL simulator"""
        for edge in edges:
            p1 = edge.FirstParameter
            p2 = edge.LastParameter
            rad1 = RadiusAt(edge, p1)
            z1 = edge.valueAt(p1).z
            if IsSame(px, rad1) and IsSame(pz, z1):
                return edge, p1, p2
            rad2 = RadiusAt(edge, p2)
            z2 = edge.valueAt(p2).z
            if IsSame(px, rad2) and IsSame(pz, z2):
                return edge, p2, p1
            # sometimes a flat circle is without edge, so return edge with
            # same height and later a connecting edge will be interpolated
            if IsSame(pz, z1):
                return edge, p1, p2
            if IsSame(pz, z2):
                return edge, p2, p1
        return None, 0.0, 0.0

    def FindTopMostEdge(self, edges):
        """Examine tool solid edges and find the top most one"""
        maxz = -99999999.0
        topedge = None
        top_p1 = 0.0
        top_p2 = 0.0
        for edge in edges:
            p1 = edge.FirstParameter
            p2 = edge.LastParameter
            z = edge.valueAt(p1).z
            if z > maxz:
                topedge = edge
                top_p1 = p1
                top_p2 = p2
                maxz = z
            z = edge.valueAt(p2).z
            if z > maxz:
                topedge = edge
                top_p1 = p2
                top_p2 = p1
                maxz = z
        return topedge, top_p1, top_p2

    def GetToolProfile(self, tool, resolution):
        """Get the edge profile of a tool solid. Basically locating the
        side edge that OCC creates on any revolved object
        """
        originalPlacement = tool.Placement
        tool.Placement = Placement(Vector(0, 0, 0), Rotation(Vector(0, 0, 1), 0), Vector(0, 0, 0))
        shape = tool.Shape
        tool.Placement = originalPlacement
        sideEdgeList = []
        for _i, edge in enumerate(shape.Edges):
            if not edge.isClosed():
                # v1 = edge.firstVertex()
                # v2 = edge.lastVertex()
                # tp = "arc" if type(edge.Curve) is Part.Circle else "line"
                sideEdgeList.append(edge)

        # sort edges as a single 3d line on the x-z plane

        # first find the topmost edge
        edge, p1, p2 = self.FindTopMostEdge(sideEdgeList)
        profile = [RadiusAt(edge, p1), edge.valueAt(p1).z]
        endrad = 0.0
        # one by one find all connecting edges
        while edge is not None:
            sideEdgeList.remove(edge)
            if isinstance(edge.Curve, Part.Circle):
                # if edge is curved, approximate it with lines based on resolution
                nsegments = int(edge.Length / resolution) + 1
                step = (p2 - p1) / nsegments
                location = p1 + step
                while nsegments > 0:
                    endrad = RadiusAt(edge, location)
                    endz = edge.valueAt(location).z
                    profile.append(endrad)
                    profile.append(endz)
                    location += step
                    nsegments -= 1
            else:
                endrad = RadiusAt(edge, p2)
                endz = edge.valueAt(p2).z
                profile.append(endrad)
                profile.append(endz)
            edge, p1, p2 = self.FindClosestEdge(sideEdgeList, endrad, endz)
            if edge is None:
                break
            startrad = RadiusAt(edge, p1)
            if not IsSame(startrad, endrad):
                profile.append(startrad)
                startz = edge.valueAt(p1).z
                profile.append(startz)

        return profile

    def Activate(self):
        """Invoke the simulator task panel"""
        self.initdone = False
        self.taskForm = CAMSimTaskUi(self)
        form = self.taskForm.form
        self.Connect(form.toolButtonPlay, self.SimPlay)
        form.sliderAccuracy.valueChanged.connect(self.onAccuracyBarChange)
        self.onAccuracyBarChange()
        self._populateJobSelection(form)
        form.comboJobs.currentIndexChanged.connect(self.onJobChange)
        self.onJobChange()
        form.listOperations.itemChanged.connect(self.onOperationItemChange)
        FreeCADGui.Control.showDialog(self.taskForm)
        self.disableAnim = False
        self.firstDrill = True
        self.millSim = CAMSimulator.PathSim()
        self.initdone = True
        self.job = self.jobs[self.taskForm.form.comboJobs.currentIndex()]
        # self.SetupSimulation()

    def _populateJobSelection(self, form):
        """Make Job selection combobox"""
        setJobIdx = 0
        jobName = ""
        jIdx = 0
        # Get list of Job objects in active document
        jobList = FreeCAD.ActiveDocument.findObjects("Path::FeaturePython", "Job.*")
        jCnt = len(jobList)

        # Check if user has selected a specific job for simulation
        guiSelection = FreeCADGui.Selection.getSelectionEx()
        if guiSelection:  #  Identify job selected by user
            sel = guiSelection[0]
            if hasattr(sel.Object, "Proxy") and isinstance(sel.Object.Proxy, PathJob.ObjectJob):
                jobName = sel.Object.Name
                FreeCADGui.Selection.clearSelection()

        # populate the job selection combobox
        form.comboJobs.blockSignals(True)
        form.comboJobs.clear()
        form.comboJobs.blockSignals(False)
        for j in jobList:
            form.comboJobs.addItem(j.ViewObject.Icon, j.Label)
            self.jobs.append(j)
            if j.Name == jobName or jCnt == 1:
                setJobIdx = jIdx
            jIdx += 1

        # Preselect GUI-selected job in the combobox
        if jobName or jCnt == 1:
            form.comboJobs.setCurrentIndex(setJobIdx)
        else:
            form.comboJobs.setCurrentIndex(0)

    def SetupSimulation(self):
        """Prepare all selected job operations for simulation"""
        form = self.taskForm.form
        self.activeOps = []
        self.numCommands = 0
        self.ioperation = 0
        for i in range(form.listOperations.count()):
            if form.listOperations.item(i).checkState() == QtCore.Qt.CheckState.Checked:
                self.firstDrill = True
                self.activeOps.append(self.operations[i])
                self.numCommands += len(self.operations[i].Path.Commands)

        self.stock = self.job.Stock.Shape
        self.busy = False

    def onJobChange(self):
        """When a new job is selected from the drop-down, update job operation list"""
        form = self.taskForm.form
        j = self.jobs[form.comboJobs.currentIndex()]
        self.job = j
        form.listOperations.clear()
        self.operations = []
        for op in j.Operations.OutList:
            if PathUtil.opProperty(op, "Active"):
                listItem = QtGui.QListWidgetItem(op.ViewObject.Icon, op.Label)
                listItem.setFlags(listItem.flags() | QtCore.Qt.ItemIsUserCheckable)
                listItem.setCheckState(QtCore.Qt.CheckState.Checked)
                self.operations.append(op)
                form.listOperations.addItem(listItem)
        if len(j.Model.OutList) > 0:
            self.baseShape = j.Model.OutList[0].Shape
        else:
            self.baseShape = None

    def onAccuracyBarChange(self):
        """Update simulation quality"""
        form = self.taskForm.form
        self.quality = form.sliderAccuracy.value()
        qualText = QtCore.QT_TRANSLATE_NOOP("CAM_Simulator", "High")
        if self.quality < 4:
            qualText = QtCore.QT_TRANSLATE_NOOP("CAM_Simulator", "Low")
        elif self.quality < 9:
            qualText = QtCore.QT_TRANSLATE_NOOP("CAM_Simulator", "Medium")
        form.labelAccuracy.setText(qualText)

    def onOperationItemChange(self, _item):
        """Check if at least one operation is selected to enable the Play button"""
        playvalid = False
        form = self.taskForm.form
        for i in range(form.listOperations.count()):
            if form.listOperations.item(i).checkState() == QtCore.Qt.CheckState.Checked:
                playvalid = True
                break
        form.toolButtonPlay.setEnabled(playvalid)

    def SimPlay(self):
        """Activate the simulation"""
        self.SetupSimulation()
        self.millSim.ResetSimulation()
        for op in self.activeOps:
            tool = PathDressup.toolController(op).Tool
            toolNumber = PathDressup.toolController(op).ToolNumber
            toolProfile = self.GetToolProfile(tool, 0.5)
            self.millSim.AddTool(toolProfile, toolNumber, tool.Diameter, 1)
            opCommands = PathUtils.getPathWithPlacement(op).Commands
            for cmd in opCommands:
                self.millSim.AddCommand(cmd)
        self.millSim.BeginSimulation(self.stock, self.quality)
        if self.baseShape is not None:
            self.millSim.SetBaseShape(self.baseShape, 1)

    def cancel(self):
        """Cancel the simulation"""


class CommandCAMSimulate:
    """FreeCAD invoke simulation task panel command"""

    def GetResources(self):
        """Command info"""
        return {
            "Pixmap": "CAM_SimulatorGL",
            "MenuText": QtCore.QT_TRANSLATE_NOOP("CAM_Simulator", "CAM Simulator"),
            "Accel": "P, N",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP("CAM_Simulator", "Simulates G-code on stock"),
        }

    def IsActive(self):
        """Command is active if at least one CAM job exists"""
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):
        """Activate the simulation"""
        CamSimulation = CAMSimulation()
        CamSimulation.Activate()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_SimulatorGL", CommandCAMSimulate())
    FreeCAD.Console.PrintLog("Loading PathSimulator Gui… done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# **************************************************************************
#   Copyright (c) 2024 Werner Mayer <wmayer[at]users.sourceforge.net>     *
#                                                                         *
#   This file is part of FreeCAD.                                         *
#                                                                         *
#   FreeCAD is free software: you can redistribute it and/or modify it    *
#   under the terms of the GNU Lesser General Public License as           *
#   published by the Free Software Foundation, either version 2.1 of the  *
#   License, or (at your option) any later version.                       *
#                                                                         *
#   FreeCAD is distributed in the hope that it will be useful, but        *
#   WITHOUT ANY WARRANTY; without even the implied warranty of            *
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
#   Lesser General Public License for more details.                       *
#                                                                         *
#   You should have received a copy of the GNU Lesser General Public      *
#   License along with FreeCAD. If not, see                               *
#   <https://www.gnu.org/licenses/>.                                      *
#                                                                         *
# **************************************************************************

import os
import tempfile
import unittest
import FreeCAD as App
import ImportGui
from pivy import coin


class ExportImportTest(unittest.TestCase):
    def setUp(self):
        TempPath = tempfile.gettempdir()
        self.fileName = TempPath + os.sep + "ColorPerFaceTest.step"
        self.doc = App.newDocument()

    def tearDown(self):
        App.closeDocument(self.doc.Name)

    def testSaveLoadStepFile(self):
        """
        Create a STEP file with color per face
        """
        part = self.doc.addObject("App::Part", "Part")
        box = part.newObject("Part::Box", "Box")
        self.doc.recompute()

        box.ViewObject.DiffuseColor = [
            (1.0, 0.0, 0.0, 1.0),
            (1.0, 0.0, 0.0, 1.0),
            (1.0, 0.0, 0.0, 1.0),
            (1.0, 0.0, 0.0, 1.0),
            (1.0, 1.0, 0.0, 1.0),
            (1.0, 1.0, 0.0, 1.0),
        ]

        ImportGui.export([part], self.fileName)

        self.doc.clearDocument()
        ImportGui.insert(name=self.fileName, docName=self.doc.Name, merge=False, useLinkGroup=True)

        part_features = list(filter(lambda x: x.isDerivedFrom("Part::Feature"), self.doc.Objects))
        self.assertEqual(len(part_features), 1)
        feature = part_features[0]

        self.assertEqual(len(feature.ViewObject.DiffuseColor), 6)
        self.assertEqual(feature.ViewObject.DiffuseColor[0], (1.0, 0.0, 0.0, 1.0))
        self.assertEqual(feature.ViewObject.DiffuseColor[1], (1.0, 0.0, 0.0, 1.0))
        self.assertEqual(feature.ViewObject.DiffuseColor[2], (1.0, 0.0, 0.0, 1.0))
        self.assertEqual(feature.ViewObject.DiffuseColor[3], (1.0, 0.0, 0.0, 1.0))
        self.assertEqual(feature.ViewObject.DiffuseColor[4], (1.0, 1.0, 0.0, 1.0))
        self.assertEqual(feature.ViewObject.DiffuseColor[5], (1.0, 1.0, 0.0, 1.0))

        sa = coin.SoSearchAction()
        sa.setType(coin.SoMaterialBinding.getClassTypeId())
        # We need an easier way to access nodes of a display mode
        sa.setInterest(coin.SoSearchAction.ALL)
        sa.apply(feature.ViewObject.RootNode)
        paths = sa.getPaths()

        bind = paths.get(1).getTail()
        self.assertEqual(bind.value.getValue(), bind.PER_PART)

        sa = coin.SoSearchAction()
        sa.setType(coin.SoMaterial.getClassTypeId())
        # We need an easier way to access nodes of a display mode
        sa.setInterest(coin.SoSearchAction.ALL)
        sa.apply(feature.ViewObject.RootNode)
        paths = sa.getPaths()

        mat = paths.get(1).getTail()
        self.assertEqual(mat.diffuseColor.getNum(), 6)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Utilities to help people verify and update their version of ifcopenshell"""

from packaging.version import Version

import FreeCAD
import FreeCADGui
from addonmanager_utilities import create_pip_call

translate = FreeCAD.Qt.translate
QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class IFC_UpdateIOS:
    """Shows a dialog to update IfcOpenShell"""

    def GetResources(self):
        tt = QT_TRANSLATE_NOOP("IFC_UpdateIOS", "Shows a dialog to update IfcOpenShell")
        return {
            "Pixmap": "IFC",
            "MenuText": QT_TRANSLATE_NOOP("IFC_UpdateIOS", "IfcOpenShell Update"),
            "ToolTip": tt,
        }

    def Activated(self):
        """Shows the updater UI"""

        version = self.get_current_version()
        avail = self.get_avail_version()
        if avail:
            if version:
                if Version(version) < Version(avail):
                    self.show_dialog("update", avail)
                else:
                    self.show_dialog("uptodate")
            else:
                self.show_dialog("install", avail)
        else:
            if version:
                self.show_dialog("uptodate")
            else:
                self.show_dialog("failed")

    def show_dialog(self, mode, version=None):
        """Shows a dialog to the user"""

        from PySide import QtGui

        title = translate("BIM", "IfcOpenShell update")
        note = translate(
            "BIM",
            "The update is installed in your FreeCAD's user directory and will not affect the rest of your system.",
        )
        if mode == "update":
            text = translate("BIM", "An update to your installed IfcOpenShell version is available")
            text += ": " + version + ". "
            text += translate("BIM", "Would you like to install that update?")
            text += " " + note
            buttons = QtGui.QMessageBox.Cancel | QtGui.QMessageBox.Ok
        elif mode == "uptodate":
            text = translate("BIM", "Your version of IfcOpenShell is already up to date")
            buttons = QtGui.QMessageBox.Ok
        elif mode == "install":
            text = translate("BIM", "No existing IfcOpenShell installation found on this system.")
            text += " "
            text += translate("BIM", "Would you like to install the most recent version?")
            text += " (" + version + ") " + note
            buttons = QtGui.QMessageBox.Cancel | QtGui.QMessageBox.Ok
        elif mode == "failed":
            text = translate(
                "BIM",
                "IfcOpenShell is not installed, and FreeCAD failed to find a suitable version to install. You can still install IfcOpenShell manually, visit https://wiki.freecad.org/IfcOpenShell for further instructions.",
            )
            buttons = QtGui.QMessageBox.Ok
        reply = QtGui.QMessageBox.information(None, title, text, buttons)
        if reply == QtGui.QMessageBox.Ok:
            if mode in ["update", "install"]:
                result = self.install()
                if result:
                    FreeCAD.Console.PrintLog(f"{result.stdout}\n")
                    text = translate("BIM", "IfcOpenShell update successfully installed.")
                    buttons = QtGui.QMessageBox.Ok
                    reply = QtGui.QMessageBox.information(None, title, text, buttons)

    def install(self):
        """Installs the given version"""

        import addonmanager_utilities as utils
        from PySide import QtCore, QtGui

        QtGui.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        vendor_path = utils.get_pip_target_directory()
        args = [
            "install",
            "--upgrade",
            "--disable-pip-version-check",
            "--target",
            vendor_path,
            "ifcopenshell",
        ]
        result = self.run_pip(args)
        QtGui.QApplication.restoreOverrideCursor()
        return result

    def run_pip(self, args):
        """Runs a pip command"""

        import addonmanager_utilities as utils
        import freecad.utils
        from subprocess import CalledProcessError

        cmd = create_pip_call(args)
        result = None
        try:
            result = utils.run_interruptable_subprocess(cmd)
        except CalledProcessError as pe:
            FreeCAD.Console.PrintError(pe.stderr)
        except Exception as e:
            text = translate("BIM", "Unable to run pip. Ensure pip is installed on your system.")
            FreeCAD.Console.PrintError(f"{text} {str(e)}\n")
        return result

    def get_current_version(self):
        """Retrieves the current ifcopenshell version"""

        import addonmanager_utilities as utils
        from packaging.version import InvalidVersion

        try:
            import ifcopenshell

            version = ifcopenshell.version
            try:
                Version(version)
            except InvalidVersion:
                FreeCAD.Console.PrintWarning(f"Invalid IfcOpenShell version: {version}\n")
                version = ""
        except:
            version = ""

        return version

    def get_avail_version(self):
        """Retrieves an available ifcopenshell version"""

        result = self.run_pip(["index", "versions", "ifcopenshell"])
        if result:
            if result.stdout and "versions" in result.stdout:
                result = result.stdout.split()
                result = result[result.index("versions:") + 1 :]
                result = [r.strip(",") for r in result]
                return result[0]  # we return the biggest
        return None


FreeCADGui.addCommand("IFC_UpdateIOS", IFC_UpdateIOS())


# >>> utils.get_pip_target_directory()
# '/home/yorik/.local/share/FreeCAD/Mod/../AdditionalPythonPackages/py311'
# >>> import freecad.utils
# >>> freecad.utils
# <module 'freecad.utils' from '/home/yorik/Apps/FreeCAD/Ext/freecad/utils.py'>
# >>> freecad.utils.get_python_exe
# <function get_python_exe at 0x7efdebf5ede0>
# >>> freecad.utils.get_python_exe()
# '/usr/bin/python3'
# ...
# >>> run_pip(["index", "versions", "ifcopenshell"])
# CompletedProcess(args=['/usr/bin/python3', '-m', 'pip', 'index', 'versions', 'ifcopenshell'], returncode=0, stdout='ifcopenshell (0.7.0.240423)\nAvailable versions: 0.7.0.240423, 0.7.0.240418, 0.7.0.240406\n', stderr='WARNING: pip index is currently an experimental command. It may be removed/changed in a future release without prior warning.\n')
# pip install --disable-pip-version-check --target vendor_path ifcopenshell
